#
#  Plex Extension Framework
#  Copyright (C) 2008-2010 Plex Development Team (James Clarke, Elan Feingold). All Rights Reserved.
#

import Framework
import xmlrpclib, cookielib, cerealizer, sys

if sys.platform == 'win32':
  import urllib2, httplib, socket
  GLOBAL_DEFAULT_TIMEOUT = socket._GLOBAL_DEFAULT_TIMEOUT
else:
  import urllib2_new as urllib2
  import httplib_new as httplib
  GLOBAL_DEFAULT_TIMEOUT = httplib._GLOBAL_DEFAULT_TIMEOUT

cerealizer.register(cookielib.Cookie)

class FrameworkTransport(xmlrpclib.Transport):
  
  def __init__(self, httpkit):
    self._httpkit = httpkit
    xmlrpclib.Transport.__init__(self, use_datetime=True)
    
  @property
  def user_agent(self):
    return self._httpkit._user_agent

class XMLRPC(object):
  
  def __init__(self, httpkit):
    self._httpkit = httpkit
    
  def Proxy(self, url, encoding=None):
    """
      Returns an :class:`xmlrpclib.ServerProxy` object for the given *url*. The framework's
      default HTTP headers will be used.
    """
    if url[0:8] == 'https://':
      raise Framework.FrameworkException('HTTPS URLs are not currently supported')
    return xmlrpclib.ServerProxy(url, FrameworkTransport(self._httpkit), encoding, use_datetime=True)
  

class HTTP(Framework.bases.BaseHTTPKit):
  
  # It'd be nicer to have separate cache times for the API and the internal components, but we have
  # to set it in the networking component so the same setting gets used by all HTTP-accessing APIs. 
  @property
  def CacheTime(self):
    """
      The default cache time (in seconds) used for all HTTP requests without a specific cache time set. By default, this value is 0.
      
      :type: float
    """
    return self._core.networking.cache_time
    
  @CacheTime.setter
  def CacheTime(self, value):
    self._core.networking.cache_time = value
    
  @property
  def Headers(self):
    """
      A dictionary containing the default HTTP headers that should be issued with requests::
      
        HTTP.Headers['User-Agent'] = 'My Plug-in'
    """
    return self._host.custom_headers
    
  #Deprecated
  def SetCacheTime(self, cacheTime):
    self._core.log.warn('The HTTP.SetCacheTime() function is deprecated. Use the HTTP.CacheTime property instead.')
    self._core.networking.cache_time = cacheTime
    
  #Deprecated
  def SetHeader(self, header, value):
    self._core.log.warn('The HTTP.SetHeader() function is deprecated. Use HTTP.Headers[] to get and set headers instead.')
    self.Headers[header] = value
    
  #Deprecated
  def SetTimeout(self, timeout):
    self._core.log.warn('The HTTP.SetTimeout() function is deprecated. Use the Network.Timeout property instead.')
    self._core.networking.default_timeout = timeout
    
  def Request(self, url, values=None, headers={}, cacheTime=None, encoding=None, errors=None, timeout=GLOBAL_DEFAULT_TIMEOUT, immediate=False, sleep=0, data=None):
    """
      Creates and returns a new :class:`HTTPRequest` object.
      
      :arg url: The URL to use for the request.
      :type url: str
      
      :arg values: Keys and values to be URL encoded and provided as the request's POST body.
      :type values: dict
      
      :arg headers: Any custom HTTP headers that should be added to this request.
      :type headers: dict
      
      :arg cacheTime: The maximum age (in second) of cached data before it should be considered invalid.
      :type cacheTime: int
      
      :arg encoding: The string encoding to use for the downloaded data. If no encoding is provided, the framework will attempt to guess the encoding.
      :type encoding: str
      
      :arg errors: The error handling method to use. If *errors* is `'strict'` (the default), a `ValueError` is raised on errors, while a value of `'ignore'` causes errors to be silently ignored, and a value of `'replace'` causes the official Unicode replacement character, U+FFFD, to be used to replace input characters which cannot be decoded.
      :type errors: str
      
      :arg timeout: The maximum amount of time (in seconds) to wait for the request to return a response before timing out.
      :type timeout: float
      
      :arg immediate: If set to ``True``, the HTTP request will be made immediately when the object is created (by default, requests are delayed until the data they return is requested).
      :type immediate: bool
      
      :arg sleep: The amount of time (in seconds) that the current thread should be paused for after issuing a HTTP request. This is to ensure undue burden is not placed on the server. If the data is retrieved from the cache, this value is ignored.
      :type sleep: float
      
      :arg data: The raw POST data that should be sent with the request. This attribute cannot be used in conjunction with *values*.
      :type data: str
    """
    if self._context:
      if values == None and data == None:
        self._context.update_cache_time(cacheTime)
      else:
        self._context._cache_time = 0
        
    if url.find(':32400/') > -1 and self._core._policy.elevated_execution == False:
      raise Framework.FrameworkException("Accessing the media server's HTTP interface is not permitted.")
    all_headers = self._add_headers(headers)
    
    return self._http_request(
      url = url,
      values = values,
      headers = all_headers,
      cacheTime = cacheTime,
      encoding = encoding,
      errors = errors,
      timeout = timeout,
      immediate = immediate,
      sleep=sleep,
      data=data,
      opener=self._opener
    )
    
  def CookiesForURL(self, url):
    """
      Returns the cookies associated with the given URL.
    """
    if self._context:
      cookie_jar = self._context.cookie_jar
    else:
      cookie_jar = None
    cookies = self._core.networking.get_cookies_for_url(url, cookie_jar=cookie_jar)
    if cookies:
      return cookies
    elif cookie_jar:
      return self._core.networking.get_cookies_for_url(url)
    
  def GetCookiesForURL(self, url):
    self._core.log.warn("HTTP.GetCookiesForURL() is deprecated - please use HTTP.CookiesForURL() instead.")
    return self.CookiesForURL(url)
    
  def SetPassword(self, url, username, password, realm=None):
    return self._core.networking.set_http_password(url, username, password, realm)
    
  def PreCache(self, url, values=None, headers={}, cacheTime=None, encoding=None, errors=None):
    """
      Instructs the framework to pre-cache the result of a given HTTP request in a background
      thread. This method returns nothing - it is designed to ensure that cached data is
      available for future calls to *HTTP.Request*.
    """
    self._core.runtime.create_thread(self._precache, txn_id=self._txn_id, url=url, values=values, headers=headers, cacheTime=cacheTime, encoding=encoding, errors=errors)
    
  def _precache(self, url, values=None, headers={}, cacheTime=None, encoding=None, errors=None):
    self.Request(url=url, values=values, headers=headers, cacheTime=cacheTime, encoding=encoding, errors=errors, immediate=True)
    
  @property
  def Cookies(self):
    """
      Returns an iterable object containing all cookies
    """
    return self._core.networking._cookie_jar
    
  def ClearCookies(self):
    """
      Clears the plug-in's HTTP cookies.
    """
    if self._context:
      cookie_jar = self._context.cookie_jar
    else:
      cookie_jar = None
    self._core.networking.clear_cookies(cookie_jar=cookie_jar)
    
  def ClearCache(self):
    """
      Clears the plug-in's HTTP cache.
    """
    self._core.networking.clear_http_cache()
    
  def RandomizeUserAgent(self, browser=None):
    """
      Sets the plug-in's user agent string to a random combination of browser & operating system versions.
      
      :arg browser: The browser to generate a user agent string for. Either ``Firefox`` or ``Safari``, or ``None`` to randomly select the browser used.
      :type browser: str
    """
    self._core.networking.randomize_user_agent(self._host, browser=browser)
  
  def _end_context(self, response_headers):
    if self._context.cache_time != None:
      if self._context.cache_time == 0:
        response_headers['Cache-Control'] = 'no-cache'
      else:
        response_headers['Cache-Control'] = 'max-age=%d' % self._context.cache_time
      
    # Create a packed cookie containing all HTTP cookies and custom headers
    cookie_list = []
    if self._context.cookie_jar != None:
      for cookie in self._context.cookie_jar:
        cookie_list.append(cookie)
      
    state_data = {}
    if len(self._context.http_headers) > 0:
      state_data["http_headers"] = self._context.http_headers
      
    if len(cookie_list) > 0:
      packed_cookies = Framework.utils.pack(cookie_list)
      response_headers['Set-Cookie'] = self._core.identifier + '=' + packed_cookies
    
  @property
  def _user_agent(self):
    if 'User-Agent' in self.Headers:
      return self.Headers['User-Agent']
    return self._core.networking.headers['User-Agent']
    
class NetworkKit(Framework.bases.BaseKit):
  def _init(self):
    self._http = HTTP(self._policy_instance)
    self._xmlrpc = XMLRPC(self._http)
    self._globals = dict(
      HTTP    = self._http,
      XMLRPC  = self._xmlrpc,
    )
    self.Timeout = Framework.utils.AttrProxy(self._core.networking, 'default_timeout')
    
  @property
  def Address(self):
    """
      Returns the server's IP address. Unless the server is connected directly to the internet,
      this will usually be a local IP address.
      
      :rtype: str
    """
    return self._core.networking.address
  
  @property
  def PublicAddress(self):
    """
      Returns the public IP address of the network the server is currently connected to.
      
      :rtype: str
    """
    return self._core.networking.http_request("http://www.plexapp.com/ip.php", cacheTime=7200).content.strip()
  
  @property
  def Hostname(self):
    """
      Returns the server's hostname.
      
      :rtype: str
    """
    return self._core.networking.hostname
    
  def Socket(self):
    """
      Creates a new *socket* object.
      
      :rtype: `socket <http://docs.python.org/library/socket.html#socket-objects>`_
    """
    return self._core.networking.socket()
    
  def SSLSocket(self):
    """
      Creates a new *SSLSocket* object. These objects operate almost identically to the *socket* objects
      returned by the above method, but support SSL connections. The only additional requirement is that
      the *do_handshake()* method is called after establishing a connection.
    """
    return self._core.networking.ssl_socket()
