#
#  Plex Extension Framework
#  Copyright (C) 2008-2010 Plex Development Team (James Clarke, Elan Feingold). All Rights Reserved.
#

import Framework
import urllib
from Framework.modelling.objects import generate_class, generate_model_interface_class, generate_model_interface_container_class
from Framework.components.runtime import callback_string, indirect_callback_string


class ObjectWithHTTPHeaders(Framework.modelling.objects.Object):
  _attribute_list = ['http_headers', 'user_agent', 'http_cookies']
  
  @classmethod
  def _convert_header_dict_to_string(cls, d):
    return '&'.join(['%s=%s' % (name, urllib.quote(value)) for name, value in d.items()])
    
  @classmethod
  def _apply_http_headers(cls, self, el):
    # Get the HTTP headers
    if hasattr(self, 'http_headers') and self.http_headers:
      header_dict = dict(self.http_headers)
    else:
      header_dict = dict()
      
    # Set old-style attributes for backwards compatibility with older clients
    if 'User-Agent' in header_dict:
      el.set('userAgent', header_dict['User-Agent'])
    if 'Cookie' in header_dict:
      el.set('httpCookies', header_dict['Cookie'])
      
    # Add old-style HTTP headers to the dictionary
    if hasattr(self, 'user_agent') and self.user_agent:
      header_dict['User-Agent'] = self.user_agent
    if hasattr(self, 'http_cookies') and self.http_cookies:
      header_dict['Cookie'] = self.http_cookies
      
    # Convert the dictionary to a HTTP header string and set the attribute on the element
    if header_dict:
      el.set('httpHeaders', ObjectWithHTTPHeaders._convert_header_dict_to_string(header_dict))
    
  def _to_xml(self, context=None):
    el = Framework.modelling.objects.Object._to_xml(self, context)
    ObjectWithHTTPHeaders._apply_http_headers(self, el)
    return el
    

class ObjectWithPOSTCallback(ObjectWithHTTPHeaders):

  @classmethod
  def _apply_post_headers(cls, self, el):
    if hasattr(self, 'post_headers') and self.post_headers:
      el.set('postHeaders', ObjectWithHTTPHeaders._convert_header_dict_to_string(self.post_headers))
      
  @classmethod
  def _apply_post_url_and_headers(cls, self, el):
    # If the key is an instance of callback_string, check for the extra POST callback
    # properties and add them to the XML if found.
    if isinstance(self, callback_string) and self.post_url:
      el.set('postURL', self.post_url)
      ObjectWithPOSTCallback._apply_post_headers(self, el)
    
  def _to_xml(self, context=None):
    el = ObjectWithHTTPHeaders._to_xml(self, context)
    ObjectWithPOSTCallback._apply_post_url_and_headers(self.key, el)
    return el
    
    
class ProviderObject(Framework.modelling.objects.Object):
  xml_tag = 'Provider'
  _attribute_list = ['key', 'title', 'type', 'machine_identifier']


class PartObject(ObjectWithPOSTCallback):
  xml_tag = 'Part'
  _attribute_list = ['key', 'file', 'duration', 'http_headers']

  def __init__(self, **kwargs):
    if 'file' not in kwargs:
      kwargs['file'] = ''
    Framework.modelling.objects.Object.__init__(self, **kwargs)

        
class MediaObject(Framework.modelling.objects.Container):
  xml_tag = 'Media'

  _children_attr_name = 'parts'
  _child_types = [PartObject]
  _attribute_list = [
    'protocols',
    'platforms',
    'bitrate',
    'aspect_ratio',
    'audio_channels',
    'audio_codec',
    'video_codec',
    'video_resolution',
    'container',
    'video_frame_rate',
    'duration',
    'width',
    'height',
    'protocol',
    'optimized_for_streaming'
  ]
  
  def __init__(self, **kwargs):
    
    def get_list(name):
      
      if name in kwargs:
        result = kwargs[name]
        del kwargs[name]
        return result
      else:
        return []
        
    self._platforms = get_list('platforms')
    self._protocols = get_list('protocols')
    
    # If the core attribute for the given name contains one value, return it
    def set_default_attr(attr_name, attr_key):
      if hasattr(self._core, 'attributes') and attr_name not in kwargs and attr_key in self._core.attributes and len(self._core.attributes[attr_key]) == 1:
        kwargs[attr_name] = self._core.attributes[attr_key][0]
    
    set_default_attr('audio_codec', Framework.core.AUDIO_CODEC_KEY)
    set_default_attr('video_codec', Framework.core.VIDEO_CODEC_KEY)
    set_default_attr('container', Framework.core.MEDIA_CONTAINER_KEY)
    
    Framework.modelling.objects.Container.__init__(self, **kwargs)
    
  @property
  def platforms(self):
    return self._platforms
    
  @property
  def protocols(self):
    return self._protocols
    
  def _to_xml(self, context=None):
    el = Framework.modelling.objects.Container._to_xml(self, context)
    
    if context:
      if len(self) > 0:
        # If the item defines a list of platforms, check that the current platform is a valid one
        if len(self.platforms) > 0:
          if context.platform in self.platforms:
            el.set('playable', '1')
          else:
            return None
      
        # If the first media part is an IndirectFunction instance, flag it as 'indirect'
        if isinstance(self._objects[0].key, indirect_callback_string):
          el.set('indirect', '1')
      
      # If the context's protocol list contains one of this item's protocols, flag the item as 'playable'
      for protocol in context.protocols:
        if protocol in self.protocols:
          playable = True
          
          # If the object defines a video codec...
          if 'video_codec' in self._attributes and self._attributes['video_codec'] != None:
            video_codec = self._attributes['video_codec']
            
            # If the codec isn't supported by the current device, flag the item as unplayable
            if video_codec not in context.video_codecs:
              playable = False
            
            # If a resolution is defined and the device's maximum resolution is less than the resolution defined by the object, flag the item as unplayable
            elif 'resolution' in context.video_codecs[video_codec] and 'video_resolution' in self._attributes and self._attributes['video_resolution'] != None and int(context.video_codecs[video_codec]['resolution']) < self._attributes['video_resolution']:
              playable = False

          # If the object defines an audio codec, and that codec isn't supported by the current device, flag the item as unplayable
          if 'audio_codec' in self._attributes and self._attributes['audio_codec'] != None and self._attributes['audio_codec'] not in context.audio_codecs:
            playable = False
              
          el.set('playable', '1' if playable else '0')
          break
  
    return el


class DirectoryObject(ObjectWithPOSTCallback):
  xml_tag = 'Directory'
  _attribute_list = ['key', 'title', 'tagline', 'summary', 'thumb', 'art', 'duration', 'http_headers']


class InputDirectoryObject(DirectoryObject):
  _attribute_list = ['key', 'title', 'tagline', 'summary', 'thumb', 'art', 'prompt', 'http_headers']
  
  def _to_xml(self, context=None):
    el = Framework.modelling.objects.Object._to_xml(self, context)
    el.set('search', '1')
    return el
    
    
class SearchDirectoryObject(InputDirectoryObject):

  def __init__(self, name=None, identifier=None, **kwargs):
    # Check that a 'key' attribute hasn't been passed
    if 'key' in kwargs:
      raise Framework.FrameworkException("The 'key' attribute for SearchDirectoryObject instances is generated automatically.")
    
    # If no identifier has been given, use the current plug-in's identifier
    if identifier == None:
      identifier = self._core.identifier
      
    if identifier not in self._core.services.search_services:
      raise Framework.FrameworkException("No search services are available for the identifier '%s'", identifier)
      
    # If no name has been given, use the first available service
    if name == None:
      name = self._core.services.search_services[identifier].keys()[0]
    
    # Generate a key & init the superclass
    key = '/system/services/search?identifier=%s&name=%s' % (identifier, urllib.quote(name))
    InputDirectoryObject.__init__(self, key=key, **kwargs)
  
    
class PopupDirectoryObject(DirectoryObject):
  _attribute_list = ['key', 'title', 'tagline', 'summary', 'thumb', 'art', 'duration']

  def _to_xml(self, context=None):
    el = DirectoryObject._to_xml(self, context)
    el.set('popup', '1')
    return el
    
    
class PrefsObject(Framework.modelling.objects.Object):
  xml_tag = 'Directory'
  _attribute_list = ['title', 'tagline', 'summary', 'thumb', 'art']

  def _to_xml(self, context=None):
    el = Framework.modelling.objects.Object._to_xml(self, context)
    el.set('key', '/:/plugins/%s/prefs' % self._core.identifier)
    el.set('settings', '1')
    return el
    
    
class MetadataObject(Framework.modelling.objects.ModelInterfaceObject):
  _attribute_list = ['url', 'http_cookies', 'user_agent', 'http_headers', 'deferred']
  
  def _to_xml(self, context=None):
    # Check that the object has valid attributes set and raise an exception if they are
    # None or an empty string.
    #
    # All objects require either a URL to be set, so the framework can automatically generate
    # unique identifiers (rating_key) and metadata URLs (key). If the URL is not provided,
    # developers must set the key and rating_key attributes manually and provide a valid
    # callback for full metadata objects.
    #
    empty = ('', None)

    if ((hasattr(self, 'url') and self.url not in empty) or \
      (hasattr(self, 'key') and self.key not in empty and hasattr(self, 'rating_key') and self.rating_key not in empty)) == False and not \
      (context != None and (context.indirect or context.request.uri.startswith('/:/plugins/'))):
      
      raise Framework.AttributeException('If no URL is provided, the key and rating_key attributes must be set.')
    
    el = Framework.modelling.objects.ModelInterfaceObject._to_xml(self, context)
    
    # If the URL attribute is defined, add some URL Service magic where required
    if hasattr(self, 'url') and self.url != None:
      url = self._attributes['url']
      
      # Get the current environment from the context, or the global environment if no context
      # is currently in use.
      if context == None:
        env = self._core.host.environment
      else:
        env = context._host.environment
        
      urlservice = env['URLService']
      mediaobject = env['MediaObject']
      partobject = env['PartObject']
      
      lookup_key = urlservice.LookupURLForMediaURL(url)
      
      # If a key isn't set, synthesize one using the URL service lookup method
      if hasattr(self, 'key') and self.key == None:
        el.set('key', lookup_key)
        
      # If a ratingKey isn't set, use the URL
      if hasattr(self, 'rating_key') == False or (hasattr(self, 'rating_key') and self.rating_key == None):
        el.set('ratingKey', self.url)
      
      # If there are no child objects, and this class expects children, add them by calling the URL service
      if len(self._objects) == 0 and len(type(self)._child_types) > 0:
        
        # If the media objects function flagged as deferred, set the deferred attribute on the metadata object
        if urlservice._media_objects_function_for_url_is_deferred(url):
          self.deferred = True
          
        # If the item is deferred, add a synthetic item for older clients
        if self.deferred:
          el.set('deferred', '1')
          self._append_children(
            el,
            [mediaobject(
              parts = [
                partobject(key = indirect_callback_string(lookup_key + "&limit=1"))
              ]
            )],
            context
          )
          
        # If the item is not deferred, add its children normally
        else:
          items = urlservice.MediaObjectsForURL(url, allow_deferred=True)
          if items:
            self._append_children(el, items, context)
          
        
          
    # If this object supports setting HTTP headers, it's a non-media object - apply HTTP headers and POST callback info
    if 'http_headers' in self._attribute_list:
      ObjectWithHTTPHeaders._apply_http_headers(self, el)
      ObjectWithPOSTCallback._apply_post_url_and_headers(self.key, el)
        
    return el
    
    
class ObjectContainer(Framework.modelling.objects.ModelInterfaceObjectContainer):
  
  _attribute_list = ['view_group', 'content', 'art', 'title1', 'title2', 'user_agent', 'http_cookies', 'no_history', 'replace_parent', 'mixed_parents', 'no_cache', 'header', 'message', 'http_headers', 'post_headers', 'identifier']
  
  def _set_attribute(self, el, name, value):
    if name == 'view_group':
      grp = self._core.runtime.view_groups[value]
      if grp.viewMode: el.set("viewmode", str(grp.viewMode))
      if grp.mediaType: el.set("contenttype", str(grp.mediaType))
      if grp.viewType: el.set("viewType", str(grp.viewType))
      if grp.viewMenu != None:
        if grp.viewMenu:
          el.set("viewMenu", "1")
        else:
          el.set("viewMenu", "0")
      if grp.viewThumb != None:
        if grp.viewThumb:
          el.set("viewThumb", "1")
        else:
          el.set("viewThumb", "0")
      if grp.viewCols != None: el.set("viewCols", str(grp.viewCols))
      if grp.viewRows != None: el.set("viewRows", str(grp.viewRows))
      if grp.viewSummary != None: el.set("viewSummary", str(grp.viewSummary))
    
    elif name == 'content':
      Framework.modelling.objects.ModelInterfaceObjectContainer._set_attribute(self, el, "viewGroup", value)

    else:
      Framework.modelling.objects.ModelInterfaceObjectContainer._set_attribute(self, el, name, value)
      
      
  def _to_xml(self, context=None):
    el = Framework.modelling.objects.ModelInterfaceObjectContainer._to_xml(self, context)
    if not (hasattr(self, 'identifier') and self.identifier != None):
      Framework.modelling.objects.ModelInterfaceObjectContainer._set_attribute(self, el, "identifier", self._core.identifier)
    
    # Set the prefix & version for media flags
    Framework.modelling.objects.ModelInterfaceObjectContainer._set_attribute(self, el, "mediaTagPrefix", "/system/bundle/media/flags/")
    sub_file_path = self._core.storage.join_path(self._core.app_support_path, "Plug-ins", "Media-Flags.bundle", "Contents", "Resources", "substitutions.xml")
    if self._core.storage.file_exists(sub_file_path):
      last_modified = self._core.storage.last_modified(sub_file_path)
      Framework.modelling.objects.ModelInterfaceObjectContainer._set_attribute(self, el, "mediaTagVersion", str(int(last_modified)))
    
    # Set default HTTP headers
    ObjectWithHTTPHeaders._apply_http_headers(self, el)
    
    # Set default headers to use for requests for POST callback data
    ObjectWithPOSTCallback._apply_post_headers(self, el)
      
    return el
      

class ObjectKit(Framework.bases.BaseKit):

  root_object = False
  
  def _generate_object(self, model, media=True):
    # If this is a media-containing object, allow MediaObject child items
    child_types = [MediaObject] if media else []
    
    # Add extra attributes for URL services
    attribute_list = ['url', 'http_cookies', 'user_agent']
    
    # Add the http_headers attribute if this is a non-media object
    if not media:
      attribute_list.append('http_headers')
    
    # Add the deferred attribute if this is a media object
    else:
      attribute_list.append('deferred')
    
    return generate_model_interface_class(self._core, model, superclass=MetadataObject, attribute_list = attribute_list, child_types = child_types, children_attr_name = 'items')

  
  def _requires_context(self):
    return True
  
    
  def _web_video_url(self, url):
    prefix = self._core.runtime.prefix_handlers.keys()[0]
    return "plex://127.0.0.1/video/:/webkit?url=%s&prefix=%s" % (urllib.quote_plus(url), prefix)
    
    
  def _rtmp_video_url(self, url, clip=None, clips=None, width=None, height=None, live=False):
    final_url = "http://www.plexapp.com/player/player.php" + \
      "?url=" + urllib.quote(url) + \
      "&live="
    if live:
      final_url += "true"
    else:
      final_url += "false"
    if clip:
      final_url += "&clip=" + urllib.quote(clip)
    if clips:
      for c in clips:
        final_url += "&clip[]=" + str(c)
    if width:
      final_url += "&width=" + str(width)
    if height:
      final_url += "&height=" + str(height)
      
    return self._web_video_url(final_url)
    
    
  def _windows_media_video_url(self, url, width=None, height=None):
    final_url = "http://www.plexapp.com/player/silverlight.php" + \
      "?stream=" + urllib.quote(url)
    if width:
      final_url += "&width=" + str(width)
    if height:
      final_url += "&height=" + str(height)
    
    return self._web_video_url(final_url)
    
  
  def _init(self):
    self._accessor = self._core._metadata_model_accessor
    self._access_point = self._accessor.get_access_point(self._core.identifier)

    self._globals = dict(
      # Old-school objects
      XMLObject             = Framework.objects.ObjectFactory(self._core, Framework.objects.XMLObject),
      XMLContainer          = Framework.objects.ObjectFactory(self._core, Framework.objects.XMLContainer),
      MediaContainer        = Framework.objects.ObjectFactory(self._core, Framework.objects.MediaContainer),
      MessageContainer      = Framework.objects.ObjectFactory(self._core, Framework.objects.MessageContainer),
      DirectoryItem         = Framework.objects.ObjectFactory(self._core, Framework.objects.DirectoryItem),
      PopupDirectoryItem    = Framework.objects.ObjectFactory(self._core, Framework.objects.PopupDirectoryItem),
      SearchDirectoryItem   = Framework.objects.ObjectFactory(self._core, Framework.objects.SearchDirectoryItem),
      InputDirectoryItem    = Framework.objects.ObjectFactory(self._core, Framework.objects.InputDirectoryItem),
      VideoItem             = Framework.objects.ObjectFactory(self._core, Framework.objects.VideoItem),
      WebVideoItem          = Framework.objects.ObjectFactory(self._core, Framework.objects.WebVideoItem),
      RTMPVideoItem         = Framework.objects.ObjectFactory(self._core, Framework.objects.RTMPVideoItem),
      WindowsMediaVideoItem = Framework.objects.ObjectFactory(self._core, Framework.objects.WindowsMediaVideoItem),
      PhotoItem             = Framework.objects.ObjectFactory(self._core, Framework.objects.PhotoItem),
      TrackItem             = Framework.objects.ObjectFactory(self._core, Framework.objects.TrackItem),
      Function              = Framework.objects.ObjectFactory(self._core, Framework.objects.Function),
      IndirectFunction      = Framework.objects.ObjectFactory(self._core, Framework.objects.IndirectFunction),
      PrefsItem             = Framework.objects.ObjectFactory(self._core, Framework.objects.PrefsItem),
      Redirect              = Framework.objects.ObjectFactory(self._core, Framework.objects.Redirect),
      ContextMenu           = Framework.objects.ObjectFactory(self._core, Framework.objects.ContextMenu),
      DataObject            = Framework.objects.ObjectFactory(self._core, Framework.objects.DataObject),
      MetadataSearchResult  = Framework.objects.ObjectFactory(self._core, Framework.objects.MetadataSearchResult),
      
      # New model interface objects
      ObjectContainer       = generate_model_interface_container_class(
        self._core,
        'MediaContainer',
        superclass = ObjectContainer,
        child_types = [
          Framework.modelling.objects.ModelInterfaceObject, 
          ProviderObject,
          DirectoryObject,
          InputDirectoryObject,
          PopupDirectoryObject,
          PrefsObject
        ]),
        
      MovieObject           = self._generate_object(self._access_point.Movie),
      VideoClipObject       = self._generate_object(self._access_point.VideoClip),
      EpisodeObject         = self._generate_object(self._access_point.Episode),
      SeasonObject          = self._generate_object(self._access_point.Season, media=False),
      TVShowObject          = self._generate_object(self._access_point.TV_Show, media=False),
      ArtistObject          = self._generate_object(self._access_point.Artist, media=False),
      AlbumObject           = self._generate_object(self._access_point.Album, media=False),
      TrackObject           = self._generate_object(self._access_point.Track),
      PhotoObject           = self._generate_object(self._access_point.Photo),
      PhotoAlbumObject      = self._generate_object(self._access_point.PhotoAlbum, media=False),
      MediaObject           = generate_class(MediaObject, self._core),
      PartObject            = generate_class(PartObject, self._core),
      ProviderObject        = generate_class(ProviderObject, self._core),
      DirectoryObject       = generate_class(DirectoryObject, self._core),
      InputDirectoryObject  = generate_class(InputDirectoryObject, self._core),
      PopupDirectoryObject  = generate_class(PopupDirectoryObject, self._core),
      PrefsObject           = generate_class(PrefsObject, self._core),
      SearchDirectoryObject = generate_class(SearchDirectoryObject, self._core),
      
      # New convenience functions
      WebVideoURL           = self._web_video_url,
      RTMPVideoURL          = self._rtmp_video_url,
      WindowsMediaVideoURL  = self._windows_media_video_url
    )
      