#
#  Plex Extension Framework
#  Copyright (C) 2008-2010 Plex Development Team (James Clarke, Elan Feingold). All Rights Reserved.
#

import Framework
import plistlib, feedparser, sys, yaml


if sys.platform == 'win32':
  import socket
  GLOBAL_DEFAULT_TIMEOUT = socket._GLOBAL_DEFAULT_TIMEOUT
else:
  import httplib_new as httplib
  GLOBAL_DEFAULT_TIMEOUT = httplib._GLOBAL_DEFAULT_TIMEOUT

class Plist(Framework.bases.BaseHTTPKit):
  
  def ObjectFromString(self, string):
    """
      Returns an object representing the given Plist-formatted string.
    """
    return plistlib.readPlistFromString(string)
    
  def ObjectFromURL(self, url, values=None, headers={}, cacheTime=None, encoding=None, errors=None, timeout=GLOBAL_DEFAULT_TIMEOUT, sleep=0):
    """
      Retrieves the content for a given HTTP request and parses it as a Plist using the above method.
      
      :arg url: The URL to retrieve content from.
      :type url: str
      
      :arg values: Values to pass as URL encoded content for a POST request.
      :type values: dict
      
      :arg headers: Custom HTTP headers to add to the request.
      :type headers: dict
      
      :arg cacheTime: The maximum age (in seconds) that cached data should still be considered valid.
      :type cacheTime: float
      
      :arg timeout: The maximum amount of time (in seconds) that the framework should wait for a response before aborting.
      :type timeout: float
      
      :arg sleep: The number of seconds the current thread should pause for if a network request was made, ensuring undue burden isn't placed on web servers. If cached data was used, this value is ignored.
      :type sleep: float
    """
    if url.find(':32400/') > -1 and self._core._policy.elevated_execution == False:
      raise Framework.FrameworkException("Accessing the media server's HTTP interface is not permitted.")
    
    if self._context:
      if values == None:
        self._context.update_cache_time(cacheTime)
      else:
        self._context._cache_time = 0
        
    all_headers = self._add_headers(headers)
    return self.ObjectFromString(self._http_request(
      url = url,
      values = values,
      headers = all_headers,
      cacheTime = cacheTime,
      encoding = encoding,
      errors = errors,
      timeout = timeout,
      immediate = True,
      sleep = sleep,
      opener = self._opener,
      txn_id = self._txn_id).content)
    
  def StringFromObject(self, obj):
    """
      Converts a given object to a Plist-formatted string representation.
    """
    return plistlib.writePlistToString(obj)



class JSON(Framework.bases.BaseHTTPKit):
  
  def ObjectFromString(self, string, encoding=None):
    """
      Converts a JSON-formatted string into a Python object, usually a dictionary.
    """
    return self._core.data.json.from_string(string, encoding)
  
  def ObjectFromURL(self, url, values=None, headers={}, cacheTime=None, encoding=None, errors=None, timeout=GLOBAL_DEFAULT_TIMEOUT, sleep=0):
    """
      Retrieves the content for a given HTTP request and parses it as JSON-formatted content using the above method.
      
      :arg url: The URL to retrieve content from.
      :type url: str
      
      :arg values: Values to pass as URL encoded content for a POST request.
      :type values: dict
      
      :arg headers: Custom HTTP headers to add to the request.
      :type headers: dict
      
      :arg cacheTime: The maximum age (in seconds) that cached data should still be considered valid.
      :type cacheTime: float
      
      :arg timeout: The maximum amount of time (in seconds) that the framework should wait for a response before aborting.
      :type timeout: float
      
      :arg sleep: The number of seconds the current thread should pause for if a network request was made, ensuring undue burden isn't placed on web servers. If cached data was used, this value is ignored.
      :type sleep: float
    """
    if url.find(':32400/') > -1 and self._core._policy.elevated_execution == False:
      raise Framework.FrameworkException("Accessing the media server's HTTP interface is not permitted.")
    
    if self._context:
      if values == None:
        self._context.update_cache_time(cacheTime)
      else:
        self._context._cache_time = 0
      
    all_headers = self._add_headers(headers)
    return self.ObjectFromString(self._http_request(
      url = url,
      values = values,
      headers = all_headers,
      cacheTime = cacheTime,
      encoding = encoding,
      errors = errors,
      timeout = timeout,
      immediate = True,
      sleep = sleep,
      opener = self._opener,
      txn_id=self._txn_id).content, encoding)
    
  def StringFromObject(self, obj):
    """
      Converts the given object to a JSON-formatted string representation.
    """
    return self._core.data.json.to_string(obj)
    
    

class RSS(Framework.bases.BaseHTTPKit):

  def FeedFromString(self, string):
    """
      Parses the given string as an RSS, RDF or ATOM feed (automatically detected).
    """
    return feedparser.parse(string)

  def FeedFromURL(self, url, values=None, headers={}, cacheTime=None, encoding=None, errors=None, timeout=GLOBAL_DEFAULT_TIMEOUT, sleep=0):
    """
      Retrieves the content for a given HTTP request and parses it as an RSS, RDF or ATOM feed using the above method.
      
      :arg url: The URL to retrieve content from.
      :type url: str
      
      :arg values: Values to pass as URL encoded content for a POST request.
      :type values: dict
      
      :arg headers: Custom HTTP headers to add to the request.
      :type headers: dict
      
      :arg cacheTime: The maximum age (in seconds) that cached data should still be considered valid.
      :type cacheTime: float
      
      :arg timeout: The maximum amount of time (in seconds) that the framework should wait for a response before aborting.
      :type timeout: float
      
      :arg sleep: The number of seconds the current thread should pause for if a network request was made, ensuring undue burden isn't placed on web servers. If cached data was used, this value is ignored.
      :type sleep: float
    """
    if url.find(':32400/') > -1 and self._core._policy.elevated_execution == False:
      raise Framework.FrameworkException("Accessing the media server's HTTP interface is not permitted.")
  
    if self._context:
      if values == None:
        self._context.update_cache_time(cacheTime)
      else:
        self._context._cache_time = 0
      
    all_headers = self._add_headers(headers)
    return self.FeedFromString(self._http_request(
      url = url,
      values = values,
      headers = all_headers,
      cacheTime = cacheTime,
      encoding = encoding,
      errors = errors,
      timeout = timeout,
      immediate = True,
      sleep = sleep,
      opener = self._opener,
      txn_id = self._txn_id).content)
    


class YAML(Framework.bases.BaseHTTPKit):
  
  def ObjectFromString(self, string):
    """
      Parses the given YAML-formatted string and returns the object it represents.
    """
    obj = yaml.load(string)

  def ObjectFromURL(self, url, values=None, headers={}, cacheTime=None, encoding=None, errors=None, timeout=GLOBAL_DEFAULT_TIMEOUT, sleep=0):
    """
      Retrieves the content for a given HTTP request and parses it as YAML-formatted content using the above method.
      
      :arg url: The URL to retrieve content from.
      :type url: str
      
      :arg values: Values to pass as URL encoded content for a POST request.
      :type values: dict
      
      :arg headers: Custom HTTP headers to add to the request.
      :type headers: dict
      
      :arg cacheTime: The maximum age (in seconds) that cached data should still be considered valid.
      :type cacheTime: float
      
      :arg timeout: The maximum amount of time (in seconds) that the framework should wait for a response before aborting.
      :type timeout: float
      
      :arg sleep: The number of seconds the current thread should pause for if a network request was made, ensuring undue burden isn't placed on web servers. If cached data was used, this value is ignored.
      :type sleep: float
    """
    if url.find(':32400/') > -1 and self._core._policy.elevated_execution == False:
      raise Framework.FrameworkException("Accessing the media server's HTTP interface is not permitted.")
    
    if self._context:
      if values == None:
        self._context.update_cache_time(cacheTime)
      else:
        self._context._cache_time = 0
    
    all_headers = self._add_headers(headers)  
    return self.ObjectFromString(self._http_request(
      url = url,
      values = values,
      headers = all_headers,
      cacheTime = cacheTime,
      encoding = encoding,
      errors = errors,
      timeout = timeout,
      immediate = True,
      sleep = sleep,
      opener = self._opener,
      txn_id = self._txn_id).content)
    
    
    
class XML(Framework.bases.BaseHTTPKit):
  def Element(self, name, text=None, **kwargs):
    """
      Returns a new XML element with the given name and text content. Any keyword arguments provided will be set as attributes.
      
      :arg name: The name of the new element.
      :type name: str
      
      :arg text: The text content of the new element.
      :type text: str
      
      :rtype: `_Element <http://lxml.de/api/lxml.etree._Element-class.html>`_
    """
    return self._core.data.xml.element(name, text, **kwargs)

  def StringFromElement(self, el, encoding='utf8'):
    """
      Converts the XML element object *el* to a string representation using the given encoding.
    """
    return self._core.data.xml.to_string(el, encoding)

  def ElementFromString(self, string, encoding=None):
    """
      Converts *string* to a HTML element object.
      
      :rtype: `_Element <http://lxml.de/api/lxml.etree._Element-class.html>`_
    """
    return self._core.data.xml.from_string(string, encoding = encoding)

  def ElementFromURL(self, url, values=None, headers={}, cacheTime=None, encoding=None, errors=None, timeout=GLOBAL_DEFAULT_TIMEOUT, sleep=0):
    """
      Retrieves the content for a given HTTP request and parses it as XML using the above method.
      
      :arg url: The URL to retrieve content from.
      :type url: str
      
      :arg values: Values to pass as URL encoded content for a POST request.
      :type values: dict
      
      :arg headers: Custom HTTP headers to add to the request.
      :type headers: dict
      
      :arg cacheTime: The maximum age (in seconds) that cached data should still be considered valid.
      :type cacheTime: float
      
      :arg timeout: The maximum amount of time (in seconds) that the framework should wait for a response before aborting.
      :type timeout: float
      
      :arg sleep: The number of seconds the current thread should pause for if a network request was made, ensuring undue burden isn't placed on web servers. If cached data was used, this value is ignored.
      :type sleep: float
    """
    if url.find(':32400/') > -1 and self._core._policy.elevated_execution == False:
      raise Framework.FrameworkException("Accessing the media server's HTTP interface is not permitted.")
    
    if self._context:
      if values == None:
        self._context.update_cache_time(cacheTime)
      else:
        self._context._cache_time = 0
    
    all_headers = self._add_headers(headers)  
    return self.ElementFromString(self._http_request(
      url = url,
      values = values,
      headers = all_headers,
      cacheTime = cacheTime,
      encoding = encoding,
      errors = errors,
      timeout=timeout,
      immediate = True,
      sleep = sleep,
      opener = self._opener,
      txn_id=self._txn_id).content, encoding=encoding)
    
  def ObjectFromString(self, string):
    """
      Parses *string* as XML-formatted content and attempts to build a Python object using the objectify library.
      
      :rtype: `ObjectifiedElement <http://lxml.de/api/lxml.objectify.ObjectifiedElement-class.html>`_
    """
    return self._core.data.xml.object_from_string(string)
    
  def StringFromObject(self, obj, encoding='utf-8'):
    """
      Attempts to create objectified XML from the given object.
    """
    return self._core.data.xml.object_to_string(obj, encoding)
    
  def ObjectFromURL(self, url, values=None, headers={}, cacheTime=None, encoding=None, errors=None, timeout=GLOBAL_DEFAULT_TIMEOUT, sleep=0):
    """
      Retrieves the content for a given HTTP request and parses it as objectified XML using the above method.
      
      :arg url: The URL to retrieve content from.
      :type url: str
      
      :arg values: Values to pass as URL encoded content for a POST request.
      :type values: dict
      
      :arg headers: Custom HTTP headers to add to the request.
      :type headers: dict
      
      :arg cacheTime: The maximum age (in seconds) that cached data should still be considered valid.
      :type cacheTime: float
      
      :arg timeout: The maximum amount of time (in seconds) that the framework should wait for a response before aborting.
      :type timeout: float
      
      :arg sleep: The number of seconds the current thread should pause for if a network request was made, ensuring undue burden isn't placed on web servers. If cached data was used, this value is ignored.
      :type sleep: float
    """
    if url.find(':32400/') > -1 and self._core._policy.elevated_execution == False:
      raise Framework.FrameworkException("Accessing the media server's HTTP interface is not permitted.")
    
    if self._context:
      if values == None:
        self._context.update_cache_time(cacheTime)
      else:
        self._context._cache_time = 0
    
    all_headers = self._add_headers(headers)  
    return self.ObjectFromString(self._http_request(
      url = url,
      values = values,
      headers = all_headers,
      cacheTime = cacheTime,
      encoding = encoding,
      errors = errors,
      timeout = timeout,
      immediate = True,
      sleep = sleep,
      opener = self._opener,
      txn_id = self._txn_id).content)
    
    
    
class HTML(Framework.bases.BaseHTTPKit):
  def Element(self, name, text=None, **kwargs):
    """
      Returns a new HTML element with the given name and text content. Any keyword arguments provided will be set as attributes.
      
      :arg name: The name of the new element.
      :type name: str
      
      :arg text: The text content of the new element.
      :type text: str
      
      :rtype: `HtmlElement <http://lxml.de/api/lxml.html.HtmlElement-class.html>`_
    """
    return self._core.data.xml.html_element(name, text, **kwargs)

  def StringFromElement(self, el, encoding='utf8'):
    """
      Converts the HTML element object *el* to a string representation using the given encoding.
    """
    return self._core.data.xml.to_string(el, encoding, 'html')

  def ElementFromString(self, string):
    """
      Converts *string* to a HTML element object.
      
      :rtype: `HtmlElement <http://lxml.de/api/lxml.html.HtmlElement-class.html>`_
    """
    return self._core.data.xml.from_string(string, isHTML=True)

  def ElementFromURL(self, url, values=None, headers={}, cacheTime=None, encoding=None, errors=None, timeout=GLOBAL_DEFAULT_TIMEOUT, sleep=0):
    """
      Retrieves the content for a given HTTP request and parses it as HTML using the above method.
      
      :arg url: The URL to retrieve content from.
      :type url: str
      
      :arg values: Values to pass as URL encoded content for a POST request.
      :type values: dict
      
      :arg headers: Custom HTTP headers to add to the request.
      :type headers: dict
      
      :arg cacheTime: The maximum age (in seconds) that cached data should still be considered valid.
      :type cacheTime: float
      
      :arg timeout: The maximum amount of time (in seconds) that the framework should wait for a response before aborting.
      :type timeout: float
      
      :arg sleep: The number of seconds the current thread should pause for if a network request was made, ensuring undue burden isn't placed on web servers. If cached data was used, this value is ignored.
      :type sleep: float
    """
    if url.find(':32400/') > -1 and self._core._policy.elevated_execution == False:
      raise Framework.FrameworkException("Accessing the media server's HTTP interface is not permitted.")
    
    if self._context:
      if values == None:
        self._context.update_cache_time(cacheTime)
      else:
        self._context._cache_time = 0
        
    all_headers = self._add_headers(headers)
    return self.ElementFromString(self._http_request(
      url = url,
      values = values,
      headers = all_headers,
      cacheTime = cacheTime,
      encoding = encoding,
      errors = errors,
      timeout = timeout,
      immediate = True,
      sleep = sleep,
      opener = self._opener,
      txn_id = self._txn_id
    ).content)
  
  
class ParseKit(Framework.bases.BaseKit):
  
  root_object = False
  
  def _init(self):
    self._globals = dict(
      JSON    = JSON(self._policy_instance),
      Plist   = Plist(self._policy_instance),
      RSS     = RSS(self._policy_instance),
      YAML    = YAML(self._policy_instance),
      XML     = XML(self._policy_instance),
      HTML    = HTML(self._policy_instance)
    )
