#
#  Plex Extension Framework
#  Copyright (C) 2008-2010 Plex Development Team (James Clarke, Elan Feingold). All Rights Reserved.
#

import Framework
import os, socket, sys

if sys.platform == 'win32':
  import urllib2
else:
  import urllib2_new as urllib2

class Exceptions(object):
  FrameworkException  = Framework.FrameworkException
  ContextException    = Framework.ContextException
  SocketError         = socket.error
  HTTPError           = urllib2.HTTPError
  URLError            = urllib2.URLError
  SocketTimeout       = socket.timeout
  

class Route(Framework.bases.BaseKit):
    
  def _connect_route_decorator(self, path, method='GET', **kwargs):
    def connect_route_decorator_inner(f):
      self.Connect(path, f, method, **kwargs)
      return f
    return connect_route_decorator_inner
    
  def _generate_route(self, f, method='GET', **kwargs):
    return self._core.runtime.generate_route(f, method, **kwargs)
    
  def __call__(self, f, method='GET', **kwargs):
    return self._generate_route(f, method, **kwargs)
    
  def Connect(self, path, f, method=['GET'], **kwargs):
    """
      Provides equivalent functionality to the *route* decorator, but allows instance methods
      of objects to be added as routes as well as unbound functions.
    """
    return self._core.runtime.connect_route(path, f, method, **kwargs)


class Request(Framework.bases.BaseKit):
  
  def _requires_context(self, context):
    return True
    
  def _warn(self, attr_info):
    self._core.log.warn('Attempting to access %s outside a request context', attr_info)
    
  @property
  def Headers(self):
    """
      A dictionary of the HTTP headers received by the plug-in for the current request.
      
      :rtype: dict
    """
    if self._context:
      return self._context.request.headers
    else:
      self._warn('request headers')
      return {}
      
  @property
  def Body(self):
    if self._context:
      return self._context.request.body
    else:
      self._warn('POST body')
      return ""
      
  @property
  def Method(self):
    if self._context:
      return self._context.request.method
    else:
      self._warn('request method')
      return ""

    
    
class Response(Framework.bases.BaseKit):

  @property
  def Headers(self):
    """
      A dictionary of keys and values that will be returned to the client as HTTP headers::
      
        HTTP.Headers['MyHeader'] = 'MyValue'
    """
    if self._context:
      return self._context.response_headers
    else:
      self._core.log.warn('Attempting to access response headers outside a request context')
      return {}
      
  @property
  def Status(self):
    """
      An integer value specifying the `HTTP status code <http://en.wikipedia.org/wiki/List_of_HTTP_status_codes>`_
      of the response.
    """
    if self._context:
      return self._context.response_status
    else:
      self._core.log.warn('Attempting to access response status outside a request context')
      return None
      
  @Status.setter
  def Status(self, status):
    if self._context:
      self._context.response_status = status
    else:
      self._core.log.warn('Attempting to access response status outside a request context')
      
  def _requires_context(self, context):
    return True
    

class Client(Framework.bases.BaseKit):

  def _requires_context(self, context):
    return (Framework.constants.header.client_platform in context.headers or Framework.constants.header.client_capabilities in context.headers or Framework.constants.header.transaction_id in context.headers)

  @property
  def Platform(self):
    """
      Reports the platform of the client currently accessing the plug-in.
      
      :returns: The client platform; one of the constants defined in :mod:`ClientPlatform`.
      :rtype: str
    """
    if self._context == None:
      self._core.log.error('Client platform information is unavailable in this context.')
      return None
    return self._context.platform

  @property
  def Protocols(self):
    """
      Reports the protocols supported by the client currently accessing the plug-in.
      
      :returns: A list of protocols supported by the client, containing constants defined in :mod:`Protocol`.
      :rtype: list
    """
    if self._context == None:
      self._core.log.error('Client protocol information is unavailable in this context.')
      return []
    return self._context.protocols
    

class Plugin(Framework.bases.BaseKit):
  
  @property
  def Identifier(self):
    """
      A read-only attribute containing the identifier of the plug-in.

      :rtype: str
    """
    return self._core.identifier
    
  def AddPrefixHandler(self, prefix, handler, name, thumb="icon-default.png", art="art-default.png", titleBar="titlebar-default.png", share=False):
    self._core.runtime.add_prefix_handler(prefix, handler, name, thumb, art, titleBar, share)
    
  def AddViewGroup(self, name, viewMode="List", mediaType="items", type=None, menu=None, cols=None, rows=None, thumb=None, summary=None):
    self._core.runtime.add_view_group(name, viewMode, mediaType, type, menu, cols, rows, thumb, summary)
    
  def _handler_decorator(self, prefix, name, thumb="icon-default.png", art="art-default.png", titleBar="titlebar-default.png", share=False):
    def handler_decorator_inner(f):
      self._core.runtime.add_prefix_handler(prefix, f, self._core.localization.local_string(name), thumb, art, titleBar, share)
      return f
    return handler_decorator_inner
    
  @property
  def Prefixes(self):
    """
      Returns a list of all prefixes currently registered by the plug-in.

    	:rtype: list
  	"""
    return self._core.runtime.prefix_handlers.keys()
    
  @property
  def ViewGroups(self):
    return dict(self._core.runtime.view_groups)
    
  def Traceback(self, msg='Traceback'):
    return self._core.traceback(msg)

  def Nice(self, value):
    """
      Alters the plug-in's 'niceness' level, which affects how system resources are allocated. The
      higher the value, the fewer resources will be given to this plug-in, allowing other plug-ins
      and applications to make use of them instead.

      The value should be between 0 (the default) and 20.

      :arg value: The level of 'niceness' to apply
      :type value: int
    """
    if sys.platform == "win32":
      return
    if (value < 0):
      value = 0
    nice_inc = value - os.nice(0)
    os.nice(nice_inc)


class Platform(Framework.bases.BaseKit):

  @property
  def HasSilverlight(self):
    """
      Reports whether the server supports playback of Silverlight video content.
      
      :rtype: bool
    """
    #TODO: Check all paths here
    return os.path.exists("/Library/Internet Plug-ins/Silverlight.plugin")
    
  @property
  def HasWebKit(self):
    """
      Reports whether the server supports video playback via the WebKit engine.
      
      :rtype: bool
    """
    webkit = self._core.get_server_attribute("webkit")
    return webkit == "1" or self.OS == 'MacOSX'

  @property
  def OS(self):
    """
      Reports the current server's operating system.
      
      :returns: The current platform; either `MacOSX`, `Windows` or `Linux`.
      :rtype: str
    """
    return self._core.runtime.os
    
  @property
  def OSVersion(self):
    """
      Reports the current server's operating system version.
      
      :returns: The version of the current platform, e.g. `10.7.0`
      :rtype: str
    """
    return self._core.get_server_attribute('platformVersion')

  @property
  def CPU(self):
    """
      Reports the current server's CPU architecture.
      
      :returns: The current CPU architecture; either `i386`, `MIPS`, `mips64` or `armv5tel`.
      :rtype: str
    """
    return self._core.runtime.cpu
    
  @property
  def MachineIdentifier(self):
    """
      Reports the current server's machine identifier.
      
      :returns: The unique identifier for the server.
      :rtype: str
    """
    
    return self._core.get_server_attribute('machineIdentifier')
    
  @property
  def ServerVersion(self):
    """
      Reports the current server's version string.

      :returns: The version the server.
      :rtype: str
    """

    return self._core.get_server_attribute('serverVersion')


class RuntimeKit(Framework.bases.BaseKit):
  
  root_object = False
  
  def _init(self):
    self._route = Route(self._policy_instance)
    self._plugin = Plugin(self._policy_instance)
    
    self._globals = dict(
      Client = Client(self._policy_instance),
      Platform = Platform(self._policy_instance),
      Callback = self._core.runtime.generate_callback_path,
      Ex = Exceptions,
      indirect = Framework.components.runtime.IndirectFunction,
      deferred = Framework.components.runtime.DeferredFunction,
    )
    
    if not isinstance(self._policy_instance, Framework.policies.ServicePolicy):
      self._globals.update(dict(
        Plugin = self._plugin,
        handler = self._plugin._handler_decorator,
        Route = self._route,
        route = self._route._connect_route_decorator,
        Request = Request(self._policy_instance),
        Response = Response(self._policy_instance),
      ))
      
      
    # If we're not creating a context (i.e. this will be the global kit object), and we're not running under the service policy, add a request handler
    if self._context == None and not isinstance(self._policy_instance, Framework.policies.ServicePolicy):
      self._core.runtime.add_private_request_handler(self._runtime_request_handler)

  def _runtime_request_handler(self, pathNouns, kwargs, context):
    if len(pathNouns) == 3 and pathNouns[2] == 'root':
      # Get a list of all handler paths
      keys = self._core.runtime.prefix_handlers.keys()
      
      # Try to find a video handler first
      for handler in keys:
        if handler.startswith('/video'):
          return Framework.objects.Redirect(self._core, handler)
          
      # No video handler found? Look for a music one
      for handler in self._core.runtime.prefix_handlers.keys():
        if handler.startswith('/music'):
          return Framework.objects.Redirect(self._core, handler)
  
      # Otherwise, return the first one in the list
      return Framework.objects.Redirect(self._core, keys[0])
      