#
#  Plex Extension Framework
#  Copyright (C) 2008-2010 Plex Development Team (James Clarke, Elan Feingold). All Rights Reserved.
#

import Framework
import plistlib, re
import urllib

class BaseService(Framework.bases.BaseKit):
  
  def _requires_context(self, context):
    return True
  
  
class URLService(BaseService):
  
  def _init(self):
    # Register the URL service request handler (for old-style requests)
    self._register_private_request_handler(self._urlservice_request_handler)
      
  def _urlservice_request_handler(self, pathNouns, kwargs, context):
    if len(pathNouns) > 1 and pathNouns[0] == 'plugins' and pathNouns[1] == self._core.identifier:
      pathNouns = pathNouns[2:]
      
    # Handle URL service function callbacks
    if len(pathNouns) > 1 and pathNouns[0] == 'urlservice_function':
      
      # Old-style requests (deprecated): Service must be selected by URL matching
      if len(pathNouns) == 3:
        # Extract arguments from the request
        url = Framework.utils.unpack(pathNouns[1])
        f_name = pathNouns[2]
        f_args = Framework.utils.unpack(kwargs['args'])
        f_kwargs = Framework.utils.unpack(kwargs['kwargs'])
        
        # Strip the extension if one is found
        pos = f_name.find('.')
        if pos > -1:
          f_name = f_name[:pos]
      
        # Call the named function and return the result
        self._core.log_debug(self._txn_id, "Calling function %s for '%s'", f_name, url)

        service = self._service_for_url(url)
        if service:
          result = self._core.services._call_named_function_in_service(f_name, service, [url], context = self._context)
          return result
      
      # New-style requests: Service identifier is passed as a pathNoun
      elif len(pathNouns) == 4:
        # Extract arguments from the request
        identifier = pathNouns[1]
        url = Framework.utils.unpack(pathNouns[2])
        f_name = pathNouns[3]
        f_args = Framework.utils.unpack(kwargs['args'])
        f_kwargs = Framework.utils.unpack(kwargs['kwargs'])
        
        # Strip the extension if one is found
        pos = f_name.find('.')
        if pos > -1:
          f_name = f_name[:pos]
        
        # Find a matching service for the given identifier
        services = self._all_services_for_url(url)
        for service in services:
          if service.identifier == identifier:
            
            # Call the named function and return the result
            self._core.log_debug(self._txn_id, "Calling function %s for '%s'", f_name, url)
            #result = self._call_named_urlservice_function_for_url_in_service(f_name, url, service, f_args, f_kwargs)
            result = self._core.services._call_named_function_in_service(f_name, service, [url], context = self._context)
            return result
            
    # Handle URL lookups
    else:
      if pathNouns == ['url', 'lookup'] and 'url' in kwargs:
        self._core.log.debug("Looking up URL '%s'", kwargs['url'])
        obj = self.MetadataObjectForURL(kwargs['url'])
        if obj:
          c = self._core.host.environment['ObjectContainer']()
          c.add(obj)
          return c
          
  def _media_objects_function_for_url_is_deferred(self, url):
    service = self._service_for_url(url)
    return self._media_objects_function_for_service_is_deferred(service)
    
  def _media_objects_function_for_service_is_deferred(self, service):
    return self._function_for_service_is_deferred(service, 'MediaObjectsForURL')

  def _function_for_service_is_deferred(self, service, f_name):
    ret = self._core.services._function_in_service_is_deferred(f_name, service, context=self._context)
    return ret
      

  def _service_for_url(self, url):
    """ Find the path to a service that can handle the given URL """
    # Iterate over service sets
    for service_identifier in self._core.services.url_services:
      service_set = self._core.services.url_services[service_identifier]
      # Iterate over each service in the set
      for service_name in service_set:
        service = service_set[service_name]
        # If the service has URL patterns defined, see if the URL matches a pattern, and return
        # the service if so
        if service.patterns:
          for pattern in service.patterns:
            if re.match(pattern, url, re.IGNORECASE):
              return service
    self._core.log_debug(self._txn_id, "No service found for URL '%s'", url)
    return None
    
  def _fallback_services(self):
    """ Returns a list of all URL services with their fallback attribute set to True """
    l = list()
    for service_identifier in self._core.services.url_services:
      service_set = self._core.services.url_services[service_identifier]
      for service_name in service_set:
        service = service_set[service_name]
        if service.fallback:
          l.append(service)
    return l
    
  def _all_services_for_url(self, url):
    """
      Returns a list containing a service matching the given url (if found) and all
      fallback services.
    """
    services = list()
    service = self._service_for_url(url)
    
    if service != None:
      services.append(service)
      self._core.log_debug(None, "Found a service matching '%s' - %s (%s)", url, service.name, service.identifier)
    else:
      self._core.log_debug(None, "No matching services found for '%s'", url)
      pass

    fallbacks = self._fallback_services()
    self._core.log.debug("There are %d fallback services", len(fallbacks))
    services.extend(fallbacks)

    return services
    
  def _metadata_object_for_url_from_service(self, url, service, add_items_automatically=True, allow_deferred=True, in_container=False):
    # Call the function in the service
    metadata = self._core.services._call_named_function_in_service('MetadataObjectForURL', service, [url], context = self._context)
    if metadata == None:
      return None
      
    # Set the 'url' attribute of the returned object to the given URL
    setattr(metadata, 'url', url)
    
    # If the media objects function is deferred, set the deferred attribute
    if self._media_objects_function_for_service_is_deferred(service):
      metadata.deferred = True
    
    # If no items were provided by the developer, and this item expects children, populate them using the service
    # Although this is also done when serializing to XML in ObjectKit, we do it here to because we already
    # have the exact service we want to use, and going through URL matching again is undesirable.
    if add_items_automatically and isinstance(metadata, Framework.modelling.objects.ModelInterfaceObject) and len(type(metadata)._child_types) > 0 and len(metadata) == 0 and (metadata.deferred == False or allow_deferred):
      result = self._media_objects_for_url_from_service(url, service, allow_deferred = allow_deferred)
      if result:
        for item in result:
          metadata.add(item)
          
    # Reset the deferred flag if deferring is allowed
    if metadata.deferred and allow_deferred:
      metadata.deferred = False

    # Set the source title if one isn't provided
    if metadata.source_title == None:
      if service.source_title:
        metadata.source_title = service.source_title
      else:
        metadata.source_title = service.name
        
    if in_container:
      container = service.host.environment['ObjectContainer']()
      container.add(metadata)
      return container

    else:
      return metadata

    
  def _media_objects_for_url_from_service(self, url, service, allow_deferred=False):
    return self._core.services._call_named_function_in_service('MediaObjectsForURL', service, [url], context = self._context, allow_deferred=allow_deferred)
    
  def _normalize_url(self, url, service):
    return self._core.services._call_named_function_in_service('NormalizeURL', service, [url], context = self._context)
    
  def _get_result_from_function_for_url(self, func, url, do_normalization = True, kwargs={}):
    if func != self._normalize_url and do_normalization:
      url = self.NormalizeURL(url)
      if url == None:
        return None
    
    # Iterate over services, calling the given function in each until one is able to return a response.
    services = self._all_services_for_url(url)
    for service in services:
      try:
        # Check for a deferred media objects function
        if 'allow_deferred' in kwargs and kwargs['allow_deferred'] == False and func == self._media_objects_for_url_from_service and self._media_objects_function_for_service_is_deferred(service):
          return None
          
        result = func(url, service, **kwargs)
        if result != None:
          return result
      except:
        self._core.log_critical(None, 'Exception calling function in URL service "%s" (%s)', service.name, str(service.identifier))
        raise
        
    return None
    
    
  def MetadataObjectForURL(self, url, add_items_automatically=True, allow_deferred=True, in_container=False, do_normalization=True):
    """
      Calls MetadataObjectForURL in the appropriate URL service and returns the result, setting
      the original URL on the object and populating media items if none were provided.
    """
    return self._get_result_from_function_for_url(self._metadata_object_for_url_from_service, url, do_normalization, dict(add_items_automatically = add_items_automatically, allow_deferred = allow_deferred, in_container = in_container))
    
    
  def MediaObjectsForURL(self, url, allow_deferred=True, do_normalization = True):
    """ Calls MediaObjectsForURL in the appropriate URL service and returns the result. """
    return self._get_result_from_function_for_url(self._media_objects_for_url_from_service, url, do_normalization, dict(allow_deferred=allow_deferred))
    
    
  def LookupURLForMediaURL(self, url):
    url = self.NormalizeURL(url)
    return '/system/services/url/lookup?url='+urllib.quote(url)
    
    
  def NormalizeURL(self, url):
    """ Calls NormalizeURL in the appropriate URL service and returns the result. If an exception is thrown, or the URL service returns None, return the original URL """
    try:
      n_url = self._get_result_from_function_for_url(self._normalize_url, url)
    except:
      self._core.log_critical(None, "Error when attempting to normalize URL '%s'", url)
      return None
    if n_url == None:
      self._core.log_warn(None, "Unable to normalize URL '%s'", url)
      n_url = url
    else:
      #self._core.log_debug(None, "Normalized URL '%s' to '%s'", url, n_url)
      pass
    return n_url
    
    
  def ServiceIdentifierForURL(self, url):
    """ Returns the identifier of a service that matches the given URL, or None if no match is found """
    service = self._service_for_url(url)
    if service == None:
      return None
    else:
      return service.identifier


class SearchService(BaseService):
  
  def Query(self, query, identifier, name):
    if name == None:
      name = self._core.services.search_services[identifier].keys()[0]
    service = self._core.services.search_services[identifier][name]
    result = self._core.services._call_named_function_in_service('Search', service, f_kwargs = dict(query=query), context = self._context)
    result.identifier = identifier
    
    # Set the source_title attribute of child objects if not set by the service
    for metadata in result.objects:
      if metadata.source_title == None:
        if service.source_title:
          metadata.source_title = service.source_title
        else:
          metadata.source_title = service.name
        
    return result
    
    
class RelatedContentService(BaseService):
  
  def RelatedContentForMetadata(self, metadata, identifier, name):
    service = self._core.services.related_content_services[identifier][name]
    result = self._core.services._call_named_function_in_service('RelatedContentForMetadata', service, f_kwargs = dict(metadata=metadata), context = self._context)
    
    # Set the source_title attribute of child objects if not set by the service
    for metadata in result.objects:
      if metadata.source_title == None:
        if service.source_title:
          metadata.source_title = service.source_title
        else:
          metadata.source_title = service.name
    
    return result
    

class ServiceKit(Framework.bases.BaseKit):
  
  root_object = False
  
  def _init(self):
    self._urlservice = URLService(self._policy_instance)
    self._searchservice = SearchService(self._policy_instance)
    self._relatedcontentservice = RelatedContentService(self._policy_instance)
    
    self._globals = dict(
      URLService    = self._urlservice,
      SearchService = self._searchservice,
      RelatedContentService = self._relatedcontentservice,
    )
    
    # Add a request handler for service callbacks and service-related notifications from the System bundle
    self._register_private_request_handler(self._services_request_handler)

  def _services_request_handler(self, pathNouns, kwargs, context):
    # Strip the first two standard internal request arguments (if present)
    if len(pathNouns) > 1 and pathNouns[0] == 'plugins' and pathNouns[1] == self._core.identifier:
      pathNouns = pathNouns[2:]
    
    # Instruction from the System bundle to reload services
    # e.g. /reloadServices
    if pathNouns == ['reloadServices']:
      self._core.log_debug(None, 'Reloading services')
      self._core.services.load_all_services()
      return ''
      
      
    # External request for a URL lookup
    # e.g. /url/lookup?url=(url)
    elif pathNouns == ['url', 'lookup'] and 'url' in kwargs:
      self._core.log.debug("Looking up URL '%s'", kwargs['url'])
      obj = self._urlservice.MetadataObjectForURL(kwargs['url'])
      if obj:
        c = self._core.host.environment['ObjectContainer']()
        c.add(obj)
        return c
        
    # Request for a resource from a service
    # e.g. /serviceResource/(service_type)/(identifier)/(resource_name)
    if len(pathNouns) == 4 and pathNouns[0] == 'serviceResource':
      service_type = pathNouns[1]
      service_identifier = pathNouns[2]
      resource_name = pathNouns[3]
      
      # Check for resource names that could cause access to files outside the resource directory
      if '..' in resource_name:
        raise Exception('Bad resource name')
      
      # Get the correct dictionary of services
      attr_name = service_type + '_services'
      if not hasattr(self._core.services, attr_name):
        self._core.log.error("No services exist for type '%s'", service_type)
        return None
      all_services = getattr(self._core.services, attr_name)
      
      # Try to find a matching service and get its path
      path = None
      for ident, services in all_services.items():
        if ident == service_identifier and len(services) > 0:
          path = services[services.keys()[0]].path
          break
      
      # If a path was found, check for a matching resource, and return it if available
      if path:
        paths = [
          self._core.storage.abs_path(self._core.storage.join_path(path, '..', '..', 'Resources', resource_name)),
          self._core.storage.abs_path(self._core.storage.join_path(path, '..', '..', '..', 'Resources', resource_name))
        ]
        for p in paths:
          if self._core.storage.file_exists(p):
            return self._core.storage.load(p)
        
    
    # Request for a function call inside a service
    # e.g. /serviceFunction/(service_type)/(identifier)/(service_name)/(func_name)?args=(packed_args)&kwargs=(packed_kwargs)&(**kwargs)
    if len(pathNouns) == 5 and pathNouns[0] == 'serviceFunction':
      # Assign pathNouns to variables
      service_type = pathNouns[1]
      service_identifier = pathNouns[2]
      service_name = pathNouns[3]
      f_name = pathNouns[4]
      
      # Strip the extension if one is found
      pos = f_name.find('.')
      if pos > -1:
        f_name = f_name[:pos]
      
      # Extract args & kwargs for the function from the querystring args
      if 'args' in kwargs:
        f_args = Framework.utils.unpack(kwargs['args'])
        del kwargs['args']
      else:
        f_args = []
        
      if 'kwargs' in kwargs:
        f_kwargs = Framework.utils.unpack(kwargs['kwargs'])
        del kwargs['kwargs']
      else:
        f_kwargs = {}
        
      # Get the correct dictionary of services
      attr_name = service_type + '_services'
      if not hasattr(self._core.services, attr_name):
        self._core.log.error("No services exist for type '%s'", service_type)
        return None
      services = getattr(self._core.services, attr_name)
      
      if service_identifier in services:
        service_set = services[service_identifier]
        if service_name in service_set:
          service = service_set[service_name]
          result = self._core.services._call_named_function_in_service(f_name, service, f_args, f_kwargs, context = context)
          return result
    
    # Request for known service test URLs
    elif len(pathNouns) in (1, 2, 3) and pathNouns[0] == 'serviceTestURLs':
      results = {}
      
      # Filter service sets based on identifier if one was provided
      if len(pathNouns) == 1:
        service_sets = self._core.services.url_services
      else:
        service_sets = {pathNouns[1]: self._core.services.url_services[pathNouns[1]]}
      
      # Iterate over the service sets
      for identifier, service_set in service_sets.items():
        result = {}
        
        # Filter services based on name if one was provided
        if len(pathNouns) <= 2:
          services = service_set
        else:
          services = {pathNouns[2]: service_set[pathNouns[2]]}
        
        # Iterate over services
        for name, service in services.items():
          
          # Get the list of test URLs defined in the plist
          test_urls = list(service.test_urls)
          
          # Extend the list with services defined in the code file, if found
          try:
            test_urls_from_code = self._core.services._call_named_function_in_service('TestURLs', service, [], {}, context = context)
            if test_urls_from_code and isinstance(test_urls_from_code, list):
              test_urls.extend(test_urls_from_code)
          except:
            pass
          
          # Add the list of URLs to the service's result dictionary
          result[name] = test_urls
          
        # Add the result dictionary to the full set of results
        results[identifier] = result
      
      # Return the results to the client
      return self._core.data.json.to_string(results)
      
      