#
#  Plex Extension Framework
#  Copyright (C) 2008-2010 Plex Development Team (James Clarke, Elan Feingold). All Rights Reserved.
#

import Framework

import os
import plistlib
import logging
import logging.handlers
import traceback
import sys
import signal
import urllib
import simplejson

import code
import policies
import components

VERSION               = "2.1.5.0"
COMPATIBLE_VERSION    = "2.0a1"

POLICY_KEY            = 'PlexPluginCodePolicy'
WHITELIST_KEY         = 'PlexPluginModuleWhitelist'
API_EXCLUSION_KEY     = 'PlexPluginAPIExclusions'
CONSOLE_LOGGING_KEY   = 'PlexPluginConsoleLogging'
LOG_LEVEL_KEY         = 'PlexPluginLogLevel'
AUDIO_CODEC_KEY       = 'PlexAudioCodec'
VIDEO_CODEC_KEY       = 'PlexVideoCodec'
MEDIA_CONTAINER_KEY   = 'PlexMediaContainer'


class LogFilter(logging.Filter):
  def filter(self, record):
    return 0
    
class LogFormatter(logging.Formatter):
  def format(self, record):
    for key in record.__dict__:
      if key[0] != '_' and isinstance(record.__dict__[key], str):
        record.__dict__[key] = uni(record.__dict__[key])
    return logging.Formatter.format(self, record)
    
class FrameworkLogger(logging.Logger):
  def findCaller(self):
      """
      Find the stack frame of the caller so that we can note the source
      file name, line number and function name.
      """
      f = logging.currentframe()
      #On some versions of IronPython, currentframe() returns None if
      #IronPython isn't run with -X:Frames.
      if f is not None:
        f = f.f_back
      rv = "(unknown file)", 0, "(unknown function)"
      while hasattr(f, "f_code"):
        co = f.f_code
        filename = os.path.normcase(co.co_filename)
        if filename == logging._srcfile or (filename.endswith('/core.py') and co.co_name.startswith('log_')) or (filename.endswith('/api/logkit.py')):
          f = f.f_back
          continue
        rv = (filename, f.f_lineno, co.co_name)
        break
      return rv

class FrameworkCore(object):
  def __init__(self, bundle_path, framework_path, config_module, plist=None):
    if bundle_path:
      self.bundle_path = bundle_path
    else:
      import tempfile
      self.bundle_path = tempfile.gettempdir()
      
    self.framework_path = framework_path
    self.config = config_module
    
    self.version = VERSION
    self.compatible_version = COMPATIBLE_VERSION
    
    self.attributes = dict()
    
    #TODO: Read from environment variable
    if self.config.root_path:
      self.app_support_path = self.config.root_path
      
      # Set defaults if not provided
      if not self.config.bundles_dir_name:
        self.config.bundles_dir_name = self.config.bundle_files_dir
      if not self.config.plugin_support_dir_name:
        self.config.plugin_support_dir_name = self.config.plugin_support_files_dir
    
    else:
      if sys.platform.find('linux') == 0 and 'PLEXLOCALAPPDATA' in os.environ:
        self.app_support_path = os.path.join(os.environ['PLEXLOCALAPPDATA'], 'Plex Media Server')
      elif sys.platform == "win32":
        if 'PLEXLOCALAPPDATA' in os.environ:
          key = 'PLEXLOCALAPPDATA'
        else:
          key = 'LOCALAPPDATA'
        self.app_support_path = os.path.join(os.environ[key], 'Plex Media Server')
      else:
        self.app_support_path = os.path.join(os.environ["HOME"], 'Library', 'Application Support', 'Plex Media Server')
        
      # Set defaults if not provided
      if not self.config.bundles_dir_name:
        self.config.bundles_dir_name = 'Plug-ins'
      if not self.config.plugin_support_dir_name:
        self.config.plugin_support_dir_name = 'Plug-in Support'
        
    self.plugin_support_path = os.path.join(self.app_support_path, self.config.plugin_support_dir_name)
    
    # Read the plist file
    if plist:
      self.plist_path = None
      bundle_plist = plist
    elif bundle_path:
      self.plist_path = os.path.join(bundle_path, 'Contents', 'Info.plist')
      bundle_plist = plistlib.readPlist(self.plist_path)
    else:
      print "ERROR: No bundle or plist provided"
      
    
    self.identifier = bundle_plist["CFBundleIdentifier"]
    if API_EXCLUSION_KEY in bundle_plist:
      self._api_exclusions = list(bundle_plist[API_EXCLUSION_KEY])
    else:
      self._api_exclusions = list()
      
    if CONSOLE_LOGGING_KEY in bundle_plist:
      self.config.console_logging = (str(bundle_plist[CONSOLE_LOGGING_KEY]) == '1')
    
    if LOG_LEVEL_KEY in bundle_plist:
      self.config.log_level = bundle_plist[LOG_LEVEL_KEY]  
    
    self.storage = components.Storage(self)
    self._configure_logging()
    
    try:
      version_path = os.path.abspath(os.path.join(self.framework_path, '..', '..', '..', 'VERSION'))
      f = open(version_path, 'r')
      self.build_info = f.read()
      f.close()
    except:
      self.build_info = 'No build information available'
      
    self.log.debug('Starting framework core - Version: %s, Build: %s', self.version, self.build_info)
    
    _whitelist = list()
    _syspaths = list()
    
    # Decide which security policy to use
    policy = policies.StandardPolicy
    
    if POLICY_KEY in bundle_plist:
      policy_name = bundle_plist[POLICY_KEY]
      policy_key = policy_name + 'Policy'
      if hasattr(policies, policy_key):
        policy = getattr(policies, policy_key)
        #TODO: Check the code signature
        self.log.debug("Using the %s security policy", policy_name.lower())
        
        # Check for a whitelist in the plist file
        if policy.allow_whitelist_extension and WHITELIST_KEY in bundle_plist:
          _whitelist.extend(bundle_plist[WHITELIST_KEY])
          
    if policy == policies.StandardPolicy:
      self.log.debug("Using the standard security policy")
    
    self._policy = policy
    
    # Copy default codec and container attributes from the plist
    [self.copy_attribute(bundle_plist, attr_name) for attr_name in (AUDIO_CODEC_KEY, VIDEO_CODEC_KEY, MEDIA_CONTAINER_KEY)]
    
    # Delete the plist dictionary
    del bundle_plist
    
    # Load the stored values dictionary
    self._values = {}
    self._values_file_path = os.path.join(self.plugin_support_path, 'Data', self.identifier, 'StoredValues')
    if os.path.exists(self._values_file_path):
      try:
        f = open(self._values_file_path, 'r')
        stored_values = simplejson.loads(f.read())
        f.close()
        self._values.update(stored_values)
      except:
        pass
    
    if bundle_path:
      self.code_path = os.path.join(bundle_path, 'Contents', 'Code')
      self.init_path = os.path.join(self.code_path, '__init__.py')
    else:
      self.code_path = None
      self.init_path = None
    
    self.loader = code.CodeLoader()

    # Initialize each of the core components
    try:
      self.log.debug("Starting runtime component.")
      self.runtime        = components.Runtime(self)
      self.log.debug("Starting caching component.")
      self.caching        = components.Caching(self)
      self.log.debug("Starting data component.")
      self.data           = components.Data(self)
      self.log.debug("Starting networking component.")
      self.networking     = components.Networking(self)
      self.log.debug("Starting localization component.")
      self.localization   = components.Localization(self)
      self.log.debug("Starting messaging component.")
      self.messaging      = components.Messaging(self)
    except:
      self.log_exception("Exception starting a core component.")

    # Get information about the current server
    self.runtime.create_thread(self.get_server_info)
    
    # Create a metadata model accessor for this plug-in
    metadata_root_path = self.config.root_path if self.config.root_path != None else self.app_support_path
    self._metadata_model_accessor = Framework.modelling.ModelAccessor(
      self,
      'metadata',
      self.storage.join_path(self.framework_path, 'Models', 'Metadata', '__init__.pym'),
      self.storage.join_path(metadata_root_path, 'Metadata')
    )
    
    # If we're loading code from a bundle, and loading external libraries is allowed, add the paths to sys.path
    if policy.allow_bundled_libraries and self.bundle_path:
      sys.path.insert(0, os.path.join(self.bundle_path, 'Contents', 'Libraries', self.runtime.os, self.runtime.cpu).encode('utf-8'))
      sys.path.insert(0, os.path.join(self.bundle_path, 'Contents', 'Libraries', 'Shared').encode('utf-8'))
    
    # Load services
    self.services_bundle_path = self.path_for_bundle(self.config.services_bundle_name, self.config.services_bundle_identifier)
    self.services = ServiceStore(self)
    
    # Create a code sandbox for the bundle code
    self.host = code.CodeHost(self, self.code_path, policy)
    
    # Add whitelisted modules to the policy instance
    for name in self.config.module_whitelist:
      if name not in self.host._policy_instance._whitelist:
        self.host._policy_instance._whitelist.append(name)
        
    if len(_whitelist) > 0:
      for name in _whitelist:
        if name in self.host._policy_instance._whitelist:
          _whitelist.remove(name)
      self.log.debug("Extending whitelist: %s", str(_whitelist))
      self.host._policy_instance._whitelist.extend(_whitelist)
      
    # Get the last known Framework version, and store the current one if different
    self.last_version = self.get_value('LastVersion')
    if self.last_version != self.version:
      self.set_value('LastVersion', self.version)
      
    self.log.debug("Finished starting framework core")
    
  def get_server_attribute(self, name):
    # Return the cached attribute if we have it already
    if name in self.attributes and self.attributes[name] != None:
      return self.attributes[name]
      
    # Otherwise, try to get it from the server
    root_xml = self.data.xml.from_string(self.networking.http_request('http://127.0.0.1:32400'))
    root_el = root_xml.xpath('//MediaContainer')[0]
    self.attributes[name] = root_el.get(name)
    if self.attributes[name] != None:
      self.log_debug(None, "Attribute '%s' set to '%s'", name, self.attributes[name])
    else:
      self.log_debug(None, "Unable to read server attribute '%s'", name)
      
    return self.attributes.get(name)
    
  def get_server_info(self):
    # Get the machine identifier and server version
    try:
      root_xml = self.data.xml.from_string(self.networking.http_request('http://127.0.0.1:32400'))
      root_el = root_xml.xpath('//MediaContainer')[0]
      self.attributes['machineIdentifier'] = root_el.get('machineIdentifier')
      self.attributes['serverVersion'] = root_el.get('version')
      self.log_debug(None, "Machine identifier is %s", self.attributes.get('machineIdentifier'))
      self.log_debug(None, "Server version is %s", self.attributes.get('serverVersion'))
    except:
      self.log_warn(None, 'Unable to retrieve the machine identifier or server version.')
    
  def get_value(self, key, default=None):
    return self._values.get(key, default)
    
  def set_value(self, key, value):
    self._values[key] = value
    json_str = self.data.json.to_string(self._values)
    self.storage.save(self._values_file_path, json_str)
    
  def copy_attribute(self, bundle_plist, attr_name):
    if attr_name in bundle_plist:
      self.attributes[attr_name] = [x.replace('.', '').lower() for x in bundle_plist[attr_name]]
    
  def load_code(self, elevated=False):
    self.log.debug("Loading plug-in code")
    try:
      if self.storage.file_exists(self.init_path):
        self.init_code = self.loader.load(self.init_path, elevated)
      else:
        self.init_code = None
      return True
    except:
      self.log_except(None, 'Exception while loading code')
      return False
      
  def log_msg(self, txn_id, func, fmt, *args, **kwargs):
    if txn_id:
      fmt = ('[T:%s] ' % str(txn_id)) + fmt
    func(fmt, *args, **kwargs)
    
  def log_debug(self, txn_id, fmt, *args, **kwargs):
    self.log_msg(txn_id, self.log.debug, fmt, *args, **kwargs)
    
  def log_info(self, txn_id, fmt, *args, **kwargs):
    self.log_msg(txn_id, self.log.info, fmt, *args, **kwargs)
    
  def log_warn(self, txn_id, fmt, *args, **kwargs):
    self.log_msg(txn_id, self.log.warn, fmt, *args, **kwargs)
    
  def log_error(self, txn_id, fmt, *args, **kwargs):
    self.log_msg(txn_id, self.log.error, fmt, *args, **kwargs)
    
  def log_critical(self, txn_id, fmt, *args, **kwargs):
    self.log_msg(txn_id, self.log.critical, fmt, *args, **kwargs)
    
  def log_except(self, txn_id, fmt, *args):
    self.log_critical(txn_id, self.traceback(fmt % tuple(args)))
    
  def log_exception(self, fmt, *args):
    self.log_except(None, fmt, *args)
    
  def log_stack(self, txn_id):
    stack = ''
    lines = traceback.format_stack()[3:-3]
    for line in lines:
      if sys.prefix not in line:
        stack += '  %s\n' % line.strip()
    self.log_debug(txn_id, "Current stack:\n" + stack)
    

  def _configure_logging(self):
    logging.basicConfig()
    logging.setLoggerClass(FrameworkLogger)
    logger = logging.getLogger()
    logger.handlers[0].addFilter(LogFilter())
    
    self.log = logging.getLogger(self.identifier)
    
    if self.config.log_level == 'Critical':
      self.log.setLevel(logging.CRITICAL)
    elif self.config.log_level == 'Error':
        self.log.setLevel(logging.ERROR)
    elif self.config.log_level == 'Warning':
      self.log.setLevel(logging.WARNING)
    elif self.config.log_level == 'Info':
      self.log.setLevel(logging.INFO)
    else:
      self.log.setLevel(logging.DEBUG)
    
    if self.config.log_file:
      log_dir = os.path.dirname(self.config.log_file)
    elif self.config.root_path:
      log_dir = os.path.join(self.config.root_path, self.config.log_files_dir)
    elif sys.platform.find('linux') == 0 and 'PLEXLOCALAPPDATA' in os.environ:
      log_dir = os.path.join(os.environ['PLEXLOCALAPPDATA'], 'Plex Media Server', 'Logs', 'PMS Plugin Logs')
    elif sys.platform == 'win32':
      if 'PLEXLOCALAPPDATA' in os.environ:
        key = 'PLEXLOCALAPPDATA'
      else:
        key = 'LOCALAPPDATA'
      log_dir = os.path.join(os.environ[key], 'Plex Media Server', 'Logs', 'PMS Plugin Logs')
    else:
      log_dir = os.path.join(os.environ['HOME'], 'Library', 'Logs', 'PMS Plugin Logs')

    log_config = dict(identifier = self.identifier, port = self.config.socket_interface_port)
    if self.config.root_path:
      log_file = os.path.join(log_dir, '%(identifier)s.%(port)d.log' % log_config)
    elif not self.config.log_file:
      log_file = os.path.join(log_dir, self.identifier+'.log')
    else:
      log_file = self.config.log_file % log_config
    
    self.storage.ensure_dirs(log_dir)
    rollover = os.path.exists(log_file)
    
    if self.config.console_logging == True:
      console_handler = logging.StreamHandler()
      console_formatter = LogFormatter('%(asctime)-15s - %(name)-32s (%(thread)x) :  %(levelname)s (%(module)s:%(lineno)d) - %(message)s')
      console_handler.setFormatter(console_formatter)
      self.log.addHandler(console_handler)
    
    file_handler = logging.handlers.RotatingFileHandler(log_file, mode='w', maxBytes=1048576, backupCount=5)
    file_formatter = LogFormatter('%(asctime)-15s (%(thread)x) :  %(levelname)s (%(module)s:%(lineno)d) - %(message)s')
    file_handler.setFormatter(file_formatter)
    
    if rollover:
      try:
        file_handler.doRollover()
      except:
        self.log_except(None, 'Exception performing logfile rollover')
        
    self.log.addHandler(file_handler)
    
  """ Dump the stack of all threads to the specified file """
  def dump_thread_stacks(self, file):
    id2name = dict([(th.ident, th.name) for th in threading.enumerate()])
    code = []
    for threadId, stack in sys._current_frames().items():
        code.append("\n# Thread: %s(%d)" % (id2name[threadId], threadId))
        for filename, lineno, name, line in traceback.extract_stack(stack):
            code.append('File: "%s", line %d, in %s' % (filename, lineno, name))
            if line:
                code.append("  %s" % (line.strip()))

    f = open(file, "w")
    f.write("\n".join(code))
    f.close()
    
  def traceback(self, msg='Traceback'):
    exceptionType, exceptionValue, exceptionTraceback = sys.exc_info()
    full_entry_list = traceback.extract_tb(exceptionTraceback)
    if not self.config.show_internal_traceback_frames:
      entry_list = []
      for entry in full_entry_list:
        if not ('/%s.bundle/' % self.config.framework_bundle_name in entry[0] or '/%s/' % self.config.framework_bundle_identifier in entry[0]):
          entry_list.append(entry)
    else:
      entry_list = full_entry_list
    traceback_str = ''.join(traceback.format_list(entry_list)) + ''.join(traceback.format_exception_only(exceptionType, exceptionValue))
    return "%s (most recent call last):\n%s" % (msg, traceback_str)
    
  def start(self):
    try:
      if self.init_code:
        self.host.execute(self.init_code)
      
      self.log.debug("Running pre-flight checks")
      for name in self.host._policy_instance.api:
        kit = self.host._policy_instance.api[name]
        if isinstance(kit, Framework.bases.BaseKit):
          kit._preflight()
        
      self.host.call_named_function('Start')
      self.log.info("Started plug-in")
      return True
    except:
      self.log_except(None, "Exception starting plug-in")
      return False
      
  def path_for_bundle(self, name, identifier):
    # Check if a cloud path exists - if not, use the local naming format
    bundles_path = self.storage.join_path(self.app_support_path, self.config.bundles_dir_name)
    cloud_path = self.storage.join_path(bundles_path, identifier)
    if self.storage.dir_exists(cloud_path):
      return cloud_path
    return self.storage.join_path(bundles_path, name + '.bundle')
      
class ServiceRecord(Framework.bases.Base):
  def __init__(self, core, path, name, identifier, source_title):
    Framework.bases.Base.__init__(self, core)
    self.name = name
    self.path = self._core.storage.join_path(path, name)
    self.identifier = identifier
    self.source_title = source_title
    self._host = None
    self._mtime = 0
    
  @property
  def host(self):
    service_code_path = self._core.storage.join_path(self.path, 'ServiceCode.pys')
    mtime = self._core.storage.last_modified(service_code_path)

    if self._host and mtime == self._mtime:
      return self._host
    
    if self._host:
      self._core.log_debug(None, "Reloading service code for %s (%s)", self.name, type(self).__name__)
    else:
      self._core.log_debug(None, "Loading service code for %s (%s)", self.name, type(self).__name__)
          
    code = self._core.loader.load(service_code_path)
    host = Framework.code.CodeHost(self._core, service_code_path, Framework.policies.ServicePolicy)
    host._custom_paths = [self._core.storage.abs_path(self._core.storage.join_path(self.path, '..', '..', 'Shared Code'))]
    
    prefs_kit = host.environment['Prefs']
    prefs_kit._identifier = self.identifier
    prefs_kit._load()
    
    # Substitute the default Callback() function with service path generators
    def callback_path_generator(function, ext=None, post_url=None, post_headers={}, **kwargs):
      return self._core.services._generate_callback_path(self, function, ext, post_url, post_headers, [], kwargs)
      
    def resource_path_generator(resource_name):
      return self._core.services._generate_resource_path(self, resource_name)

    host.environment['Callback'] = callback_path_generator
    host.environment['R'] = resource_path_generator

    host.execute(code)
    
    self._host = host
    self._mtime = mtime
    
    return host

class URLServiceRecord(ServiceRecord):
  def __init__(self, core, path, name, identifier, patterns, fallback=False, source_title=None, test_urls=[]):
    ServiceRecord.__init__(self, core, path, name, identifier, source_title)
    self.patterns = patterns
    self.fallback = fallback
    self.test_urls = test_urls
    
class SearchServiceRecord(ServiceRecord):
  def __init__(self, core, path, name, identifier, source_title=None):
    ServiceRecord.__init__(self, core, path, name, identifier, source_title)
    
class RelatedContentServiceRecord(ServiceRecord):
  def __init__(self, core, path, name, identifier, source_title=None):
    ServiceRecord.__init__(self, core, path, name, identifier, source_title)
    
class ServiceStore(Framework.bases.Base):
  def _init(self):
    self._services_loaded = self._core.runtime.event()
    self._core.runtime.create_thread(self.load_all_services)
    
  def get_services_from_bundle(self, bundle_path, plist=None):
    if plist == None:
      plist = plistlib.readPlist(self._core.storage.join_path(bundle_path, 'Contents', 'Info.plist'))
    bundle_identifier = plist['CFBundleIdentifier']
    
    old_services = dict()
    services = dict()
    service_sets = dict()
    
    # Load the plist from the bundle
    self._core.log_debug(None, "  -> Loading services from '%s'", bundle_path)
    
    # Get the separate service directories
    service_dir = self._core.storage.join_path(bundle_path, "Contents", "Services")
    service_set_dir = self._core.storage.join_path(bundle_path, "Contents", "Service Sets")
    
    # Needed to convert plistlib._InternalDict types to regular dictionaries, so they can be seralised correctly
    def convert_dicts(d):
      f = {}
      for k in d:
        v = d[k]
        f[k] = convert_dicts(v) if isinstance(v, dict) else v
      return f
    
    
    # Get the list of old-style services from the main plist
    def get_old_services(service_key):
      plist_key = 'Plex%sServices' % service_key
      if plist_key in plist:
        dct = convert_dicts(plist[plist_key])
        for service_name, service in dct.items():
          
          # Check for very-old-style keys and rewrite them using the slightly newer format.
          # Makes the service dictionaries fully compatible with new-style services, requiring
          # only one dict parsing function.
          #
          if plist_key == 'PlexURLServices':
            if 'LinkedPlugin' in service:
              service['Identifier'] = service['LinkedPlugin']
              del service['LinkedPlugin']
            if 'URLPattern' in service:
              url_patterns = service.get('URLPatterns', [])
              url_patterns.append(service['URLPattern'])
              service['URLPatterns'] = url_patterns
              del service['URLPattern']
            
          if plist_key == 'PlexSearchServices':
            if isinstance(service, basestring):
              dct[service_name] = {'Identifier' : service}
            
        old_services[service_key] = dct
  
    [get_old_services(key) for key in ('URL', 'Search', 'RelatedContent')]

    def get_services(key, pl=None):
      if pl == None:
        pl = plist
      if key not in pl:
        return {}
      return convert_dicts(pl[key])
    
    def load_services_from_plist(service_plist_path, dct):
      if self._core.storage.file_exists(service_plist_path):
        # Load the plist and see what services are supported
        self._core.log.debug("      - Loading plist from '%s'", service_plist_path)
        service_plist = plistlib.readPlistFromString(self._core.storage.load(service_plist_path))
        for key in ('URL', 'Search', 'RelatedContent'):
          dct[key] = get_services(key, service_plist)
    
    # Load services plist
    service_plist_path = self._core.storage.join_path(service_dir, "ServiceInfo.plist")
    load_services_from_plist(service_plist_path, services)
    
    # Load service set plists
    if self._core.storage.dir_exists(service_set_dir):
      for name in self._core.storage.list_dir(service_set_dir):
        if name not in service_sets:
          service_sets[name] = {}
        service_plist_path = self._core.storage.join_path(service_set_dir, name, "ServiceInfo.plist")
        load_services_from_plist(service_plist_path, service_sets[name])
        
    return {
      bundle_identifier: {
        'Path': bundle_path,
        'OldServices': old_services,
        'Services': services,
        'ServiceSets': service_sets,
      }
    }
    
    
  def load_all_services(self):
    self._url_services = {}
    self._search_services = {}
    self._related_content_services = {}
    
    # Get the services dict from the system bundle
    try:
      
      # If a path to the services bundle was set in the config file, load services from the current bundle and the specified path
      if self._core.config.services_bundle_path:
        self._core.log_debug(None, 'Services path specified - loading services directly from bundles')
        services_dict = dict()
        
        # Load services from the current bundle and from Services.bundle
        services_dict.update(self.get_services_from_bundle(self._core.bundle_path))
        services_dict.update(self.get_services_from_bundle(self._core.config.services_bundle_path))
        
      # If the plug-in is daemonized, only load services from the current bundle and Services.bundle
      elif self._core.config.daemonized:
        self._core.log_debug(None, 'Plug-in is daemonized - loading services directly from bundles')
        services_dict = dict()
        
        # Load services from the current bundle and from Services.bundle
        services_dict.update(self.get_services_from_bundle(self._core.bundle_path))
        services_dict.update(self.get_services_from_bundle(self._core.storage.join_path(self._core.app_support_path, 'bundle', 'com.plexapp.system.services')))
          
      # Otherwise, get the list of services from System.bundle
      else:
        services_dict = self._core.messaging.call_external_function(
          '..system',
          '_BundleService:AllServices'
        )
        
      if services_dict == None:
        services_dict = {}
        
      for identifier in services_dict:
        try:
          service_dict = services_dict[identifier]
          path = service_dict['Path']
          services = service_dict.get('Services', {})
          service_sets = service_dict.get('ServiceSets', {})
          old_services = service_dict.get('OldServices', {})
        
          # Helper function for adding a service in a safe manner
          def add_service(root_dict, service_identifier, s_class, s_name, path_components, **kwargs):
            # Make sure a service set exists in the root dictionary for this identifier
            if service_identifier not in root_dict:
              root_dict[service_identifier] = {}
            
            try:
              root_dict[service_identifier][s_name] = s_class(
                self._core,
                self._core.storage.join_path(path, 'Contents', *path_components),
                s_name,
                service_identifier,
                **kwargs
              )
              self._core.log.debug("Loaded %s service '%s' in %s from bundle %s", str(path_components), s_name, service_identifier, identifier)
            
            except:
              self._core.log_except(None, "Error loading %s service '%s' in %s from bundle %s", str(path_components), s_name, service_identifier, identifier)
        
          def load_service_dict(dct, service_identifier, path_components, old_style=False):
            for service_name, service in dct.get('URL', {}).items():
              if old_style:
                service_identifier = service['Identifier']
              add_service(
                self._url_services,
                service_identifier,
                URLServiceRecord,
                service_name,
                path_components + (['URL Services'] if old_style else ['URL']),
                patterns = service.get('URLPatterns', list()),
                fallback = service.get('Fallback', False),
                source_title = service.get('SourceTitle', None),
                test_urls = service.get('TestURLs', [])
              )
            
            for service_name, service in dct.get('Search', {}).items():
              if old_style:
                service_identifier = service['Identifier']
              add_service(
                self._search_services,
                service_identifier,
                SearchServiceRecord,
                service_name,
                path_components + (['Search Services'] if old_style else ['Search']),
                source_title = service.get('SourceTitle', None)
              )
            
            for service_name, service in dct.get('RelatedContent', {}).items():
              if old_style:
                service_identifier = service['Identifier']
              add_service(
                self._related_content_services,
                service_identifier,
                'Related Content',
                RelatedContentServiceRecord,
                service_name,
                path_components + (['Related Content Services'] if old_style else ['Related Content']),
                source_title = service.get('SourceTitle', None)
              )
            
          load_service_dict(services, identifier, ['Services'])
          [load_service_dict(dct, service_identifier, ['Service Sets', service_identifier]) for service_identifier, dct in service_sets.items()]
          load_service_dict(old_services, identifier, [], old_style=True)
        
        except:
          self._core.log_except(None, "Error loading services from '%s'", identifier)
      
      self._core.log.debug('Loaded services')

    except:
      self._core.log_except(None, "Error loading services")
    
    finally:
      self._services_loaded.set()
    
  @property
  def url_services(self):
    self._services_loaded.wait()
    return self._url_services
    
  @property
  def search_services(self):
    self._services_loaded.wait()
    return self._search_services
    
  @property
  def related_content_services(self):
    self._services_loaded.wait()
    return self._related_content_services
    
  def _type_for_service(self, service):
    if isinstance(service, URLServiceRecord):
      service_type = 'url'
    elif isinstance(service, SearchServiceRecord):
      service_type = 'search'
    elif isinstance(service, RelatedContentServiceRecord):
      service_type = 'related_content'
    else:
      service_type = 'unknown'
    return service_type
        
  def _generate_callback_path(self, service, function, ext, post_url, post_headers, args, kwargs):
    """ Generate a callback path for a function inside a service """
    # Check for an IndirectFunction and extract the actual function from the object if necessary
    if isinstance(function, Framework.components.runtime.IndirectFunction):
      function = function._f
      indirect = True
    else:
      indirect = False
      
    service_type = self._type_for_service(service)
    
    # Format the extension, if provided
    if ext and ext[0] != '.':
      ext = '.'+ext
    elif ext == None:
      ext = ''

    path = "/:/plugins/%s/serviceFunction/%s/%s/%s/%s%s?args=%s&kwargs=%s" % (
      self._core.identifier,
      service_type,
      str(service.identifier),
      urllib.quote(str(service.name)),
      function.__name__,
      ext,
      Framework.utils.pack(args),
      Framework.utils.pack(kwargs)
    )
    
    # Convert to a callback_string object
    if indirect:
      path = Framework.components.runtime.indirect_callback_string(path + "&indirect=1")
    else:
      path = Framework.components.runtime.callback_string(path)
      
    # Add POST args
    path.post_url = post_url
    path.post_headers = post_headers
      
    return path
    
  def _generate_resource_path(self, service, resource_name):
    return "/:/plugins/%s/serviceResource/%s/%s/%s" % (
      self._core.identifier,
      self._type_for_service(service),
      str(service.identifier),
      urllib.quote(resource_name)
    )
    
  def _create_service_context(self, host, context):
    # If a context is provided, create a new context linked to this host, otherwise use the
    # global context
    if context:
      new_context = host.build_context(context.request)
      new_context.cached_http_responses = dict(context.cached_http_responses)
      new_context.cookie_jar = context.cookie_jar
    else:
      new_context = None
    return new_context
    
  def _function_in_service_is_deferred(self, fname, service, context=None):
    
    return service.host.function_is_deferred(fname, self._create_service_context(service.host, context))
    
  def _call_named_function_in_service(self, fname, service, f_args=[], f_kwargs={}, context=None, allow_deferred=False, raise_exceptions=True):
    """ Call a named function in the given service with the given arguments and return the result """
    host = service.host
    new_context = self._create_service_context(host, context)
    
    result = host.call_named_function(fname, context=new_context, allow_deferred=allow_deferred, raise_exceptions=raise_exceptions, args=f_args, kwargs=f_kwargs)
    return result
    
    
