#
#  Plex Plug-in Framework
#  Copyright (C) 2008-2010 Plex, Inc. (James Clarke, Elan Feingold). All Rights Reserved.
#

from systemservice import SystemService

class PluginInfo(object):
  
  def __init__(self, host_name, friendly_name, identifier, root, key, title, version=None, thumb=None, art=None, titleBar=None, lastAccessedAt=None, hasPrefs='0', hasStoreServices='0', share='0', local=False, platforms="", protocols=""):
    self.host_name = host_name
    self.friendly_name = friendly_name
    self.identifier = identifier
    self.prefix = ('/%s/%s' % (root, key)) if root else key
    self.title = title
    self.version = version
    self.thumb = thumb
    self.art = art
    self.titleBar = titleBar
    self.hasPrefs = (hasPrefs == '1')
    self.hasStoreServices = (hasStoreServices == '1')
    self.share = (share == '1')
    self.local = local
    self.platforms = platforms.split(',') if len(platforms) > 0 else []
    
    
  def directory_item(self):
    return DirectoryItem(
      key = String.Encode("http://%s:32400%s" % (self.host_name, self.prefix)),
      title = self.title,
      host = self.host_name,
      serverName = self.friendly_name,
      serverVersion = self.version,
      path = "%s" % self.prefix,
      art = self.art,
      thumb = self.thumb,
      titleBar = self.titleBar,
      hasPrefs = self.hasPrefs,
      hasStoreServices = self.hasStoreServices,
      local = self.local,
      share = self.share,
      identifier = self.identifier,
      type = "plugin"
    )
    
  
class LibrarySectionInfo(object):
  
  def __init__(self, host_name, friendly_name, key, type, title, machine_identifier=None, version=None, thumb=None, art=None, local=False, **kwargs):
    self.host_name = host_name
    self.friendly_name = friendly_name
    self.machine_identifier = machine_identifier
    self.key = key
    self.section_type = type
    self.title = title
    self.version = version
    self.thumb = thumb
    self.art = art
    self.share = True
    self.local = local
    self.platforms = []
    
    
  def directory_item(self):
    return DirectoryItem(
      key = String.Encode("http://%s:32400/library/sections/%s" % (self.host_name, self.key)),
      title = self.title,
      type = self.section_type,
      host = self.host_name,
      serverName = self.friendly_name,
      serverVersion = self.version,
      machineIdentifier = self.machine_identifier,
      path = "/library/sections/%s" % self.key,
      art = self.art,
      thumb = self.thumb,
      local = self.local
    )


    
class MediaServerInfo(object):
  
  def __init__(self, name, host, machineIdentifier=None, version=None, **kwargs):
    self.friendly_name = name
    self.host_name = host
    self.identifier = machineIdentifier
    self.version = version
    
    root_xml = XML.ElementFromURL('http://127.0.0.1:32400/', cacheTime=0)
    local_identifier = root_xml.get('machineIdentifier')
    self.local = (self.identifier == local_identifier)

    
  def update_server_info(self):
    Log.Debug("Updating server info for %s (%s)", self.identifier, self.friendly_name)
    self.library_sections = list()
    self.video_plugins = list()
    self.music_plugins = list()
    self.picture_plugins = list()
    self.application_plugins = list()
    self.channels = list()

    # We make this one timeout fairly quickly because (a) it should be super fast and (b) we don't want to hang in here.
    sections_xml = HTTP.Request("http://%s:32400/library/sections" % self.host_name, cacheTime=0, timeout=5)
    sections_el = XML.ElementFromString(sections_xml)

    for section_el in sections_el.xpath('//Directory'):
      key = section_el.get('key')
      if key and not (key.startswith('queues/') or key == 'shared'):
        info = LibrarySectionInfo(
          host_name = self.host_name,
          friendly_name = self.friendly_name,
          version = self.version,
          machine_identifier = self.identifier,
          local = self.local,
          **dict(section_el.attrib)
        )
        Log.Debug(" -> Adding library section %s - %s", info.key, info.title)
        self.library_sections.append(info)
        
    # Try to hit /channels/all first
    try:
      # There appears to be a PMS bug that causes elements to be missing their identifiers
      # at first. This mini-hack is needed for now to work around that.
      complete = False
      count = 0
      
      while not complete:
        channels_el = XML.ElementFromURL("http://%s:32400/channels/all" % self.host_name, cacheTime=0, timeout=10)
        
        try:
          for el in channels_el.xpath('//Directory'):
            info = PluginInfo(
              host_name = self.host_name,
              friendly_name = self.friendly_name,
              version = self.version,
              root = None,
              local = self.local,
              **dict(el.attrib)
            )
            Log.Debug(" -> Adding channel %s (%s)", info.title, info.prefix)
            self.channels.append(info)
          complete = True
        
        except:
          complete = False
          count += 1
          Log.Exception("Error adding channels from /channels/all")
          if count == 5:
            raise # Re-raise so the old code below runs
          Thread.Sleep(0.5)
        
    # If the server doesn't support /channels/all, fall back to the old-school way of doing things
    except:
      video_el = XML.ElementFromURL("http://%s:32400/video" % self.host_name, cacheTime=0, timeout=10)
      for el in video_el.xpath('//Directory'):
        info = PluginInfo(
          host_name = self.host_name,
          friendly_name = self.friendly_name,
          version = self.version,
          root = 'video',
          local = self.local,
          **dict(el.attrib)
        )
        Log.Debug(" -> Adding video channel %s", info.title)
        self.video_plugins.append(info)

      music_el = XML.ElementFromURL("http://%s:32400/music" % self.host_name, cacheTime=0, timeout=10)
      for el in music_el.xpath('//Directory'):
        info = PluginInfo(
          host_name = self.host_name,
          friendly_name = self.friendly_name,
          version = self.version,
          root = 'music',
          local = self.local,
          **dict(el.attrib)
        )
        Log.Debug(" -> Adding music channel %s", info.title)
        self.music_plugins.append(info)

      picture_el = XML.ElementFromURL("http://%s:32400/photos" % self.host_name, cacheTime=0, timeout=10)
      for el in picture_el.xpath('//Directory'):
        info = PluginInfo(
          host_name = self.host_name,
          friendly_name = self.friendly_name,
          version = self.version,
          root = 'photos',
          local = self.local,
          **dict(el.attrib)
        )
        Log.Debug(" -> Adding photo channel %s", info.title)
        self.picture_plugins.append(info)
      
      application_el = XML.ElementFromURL("http://%s:32400/applications" % self.host_name, cacheTime=0, timeout=10)
      for el in application_el.xpath('//Directory'):
        info = PluginInfo(
          host_name = self.host_name,
          friendly_name = self.friendly_name,
          version = self.version,
          root = 'applications',
          local = self.local,
          **dict(el.attrib)
        )
        Log.Debug(" -> Adding application channel %s", info.title)
        self.application_plugins.append(info)
      

      
      
      
class PeerService(SystemService):

  def __init__(self, system):
    SystemService.__init__(self, system)
    Log.Debug("Starting the peer service")
    
    self.servers = dict()
    self.needs_update = True
    
    Route.Connect("/system/notify/serverUpdated", self.notify_server_updated)
    Route.Connect("/system/library", self.library_root_menu)
    Route.Connect("/system/library/sections", self.library_sections)
    Route.Connect("/system/plugins", self.plugins_root_menu)
    Route.Connect("/system/plugins/all", self.all_plugins)
    Route.Connect("/system/plugins/video", self.video_plugins)
    Route.Connect("/system/plugins/music", self.music_plugins)
    Route.Connect("/system/plugins/photos", self.picture_plugins)
    Route.Connect("/system/plugins/applications", self.application_plugins)
    Route.Connect("/system/search", self.search_providers)
    
    
  def library_root_menu(self):
    c = MediaContainer()
    c.Append(DirectoryItem('sections', 'Library Sections'))
    return c
    
    
  def plugins_root_menu(self):
    c = MediaContainer()
    c.Append(DirectoryItem('all', 'All Plug-ins'))
    c.Append(DirectoryItem('music', 'Music Plug-ins'))
    c.Append(DirectoryItem('photos', 'Picture Plug-ins'))
    c.Append(DirectoryItem('video', 'Video Plug-ins'))
    c.Append(DirectoryItem('applications', 'Application Plug-ins'))
    return c
    
  def notify_server_updated(self, host=None):
    if not host:
      host = Network.Hostname
    
    orig_host = host
    host = Core.networking.resolve_hostname_if_required(host)
    Log.Debug('Server was updated: %s (resolved to %s)' % (orig_host, host))
    
    servers_xml = HTTP.Request("http://127.0.0.1:32400/servers", cacheTime=0)
    servers_el = XML.ElementFromString(servers_xml)
    
    # Look for the server in the list.
    xpath_results = servers_el.xpath('//Server[@host="%s"]' % orig_host)
    if len(xpath_results) == 0 and orig_host.endswith('.local'):
      xpath_results = servers_el.xpath('//Server[@host="%s"]' % (orig_host+'.'))
    
    # See if PMS knows about the server.
    if len(xpath_results) == 0:
      Log.Debug("No server elements found for host %s, perhaps it went away", orig_host)
      identifier = None
    else:
      identifier = xpath_results[0].get('machineIdentifier')
    
    # Update our own list.
    self.update_server_info(identifier_to_update = identifier)
    return ''

  def plugin_container(self, *attr_names):
    Log.Info("Constructing a plug-in container for client '%s'" % str(Client.Platform))

    if self.needs_update or (Platform.MachineIdentifier != None and Platform.MachineIdentifier not in self.servers):
      self.update_server_info()
      
    plugins = dict()
    
    @lock(self.servers)
    def servers_lock():
      
      # Iterate over the server list
      for name in self.servers:
        
        # Collect items from each server for each attribute name passed in
        items = list()
        for attr_name in attr_names:
          if hasattr(self.servers[name], attr_name):
            info_list = getattr(self.servers[name], attr_name)
            
            # Check each item in the list - if it's local, or shared, add it to the aggregated list
            for info in info_list:
              if info.local or info.share:
                # First, check if any platforms are defined, and compare it to the current client platform
                if len(info.platforms) == 0 or Client.Platform == None or 'X-Plex-Disable-Capability-Checking' in Request.Headers or Client.Platform in info.platforms:
                  items.append(info.directory_item())
                
        # Iterate over the collected items & create a dictionary based on their names
        for item in items:
          if item.name not in plugins:
            plugins[item.name] = list()
          plugins[item.name].append(item)
          
    # Create a MediaContainer from the available items and return it
    c = MediaContainer()
    for name in sorted(plugins.keys(), key = lambda x: x.lower()):
      unique = (len(plugins[name]) == 1)
      for item in plugins[name]:
        item.unique = unique
        c.Append(item)
    return c

    
  def library_sections(self):
    return self.plugin_container('library_sections')

    
  def all_plugins(self, refresh=False):
    container = self.plugin_container('music_plugins', 'video_plugins', 'picture_plugins', 'application_plugins', 'channels')
    store_plugins = self.system.storeservice.plugins
    
    # Check for plug-ins managed by the store that don't report it themselves and override the attribute
    for d in container:
      if not d.hasStoreServices and d.identifier in store_plugins:
        d.hasStoreServices = True
    
    return container
        
        
  def music_plugins(self):
    return self.plugin_container('music_plugins')

  
  def picture_plugins(self):
    return self.plugin_container('picture_plugins')


  def video_plugins(self):
    return self.plugin_container('video_plugins')


  def application_plugins(self):
    return self.plugin_container('application_plugins')


  def update_server_info(self, identifier_to_update=None):
    Log.Debug("Refreshing list of servers with identifier_to_update=%s." % identifier_to_update)
    
    # Get a list of known servers from ourself.
    servers_xml = HTTP.Request("http://127.0.0.1:32400/servers", cacheTime=0)
    servers_el = XML.ElementFromString(servers_xml)
    
    @lock(self.servers)
    def servers_lock():
      
      old_identifiers = self.servers.keys()
      
      # For each currently known server.
      for server_el in servers_el.xpath('//Server'):
        try:
          # Build a host object from the XML.
          attrs = dict(server_el.attrib)
          host = Core.networking.resolve_hostname_if_required(attrs['host'])
          identifier = attrs["machineIdentifier"]
          attrs['host'] = host
          
          # Whack the old one if it existed.
          if identifier in old_identifiers:
            old_identifiers.remove(identifier)
            
          # Either build a new or update an existing server. FIXME, what if 
          # machineIdentifier changed, e.g. (or version)?
          #
          if identifier not in self.servers:
            Log.Debug("Creating a server record for %s", host)
            info = MediaServerInfo(**attrs)
            self.servers[identifier] = info
          else:
            Log.Debug("Updating server record for %s", host)
            self.servers[identifier].friendly_name = attrs['name']
            
          # Finally, go out and get information for the server.
          if identifier_to_update == None or identifier_to_update == identifier:
            self.servers[identifier].update_server_info()
          
        except:
          Core.log_except(None, "Exception getting server information")
        
      for identifier in old_identifiers:
        Log.Debug("Whacking server that went away: %s", self.servers[identifier].friendly_name)
        del self.servers[identifier]
    
      Log.Debug("Server list:")
      for identifier in self.servers:
        Log.Debug(" -> %s", identifier)
        
    self.needs_update = False
    
  def search_providers(self, query=None):
    if self.needs_update:
      self.update_server_info()
    
    c = ObjectContainer()
    @lock(self.servers)
    def servers_lock():
      for identifier in self.servers:
        server = self.servers[identifier]
        if not server.local:
          c.add(ProviderObject(
            title = server.friendly_name,
            key = 'http://%s:32400/search?local=1' % Core.networking.resolve_hostname_if_required(server.host_name),
            type = 'mixed',
            machine_identifier = server.identifier
          ))
    
    if query != None and len(query) > 2:
      for identifier in Core.services.search_services:
        for name in Core.services.search_services[identifier]:
          c.add(ProviderObject(
            title = name,
            key = '/system/services/search?identifier=%s&name=%s' % (identifier, String.Quote(name)),
            type = 'mixed'
          ))
      
    return c